/* Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

// CK_DEF.H

#include "ID_HEADS.H"
#include <BIOS.H>

#define FRILLS 1 // Bringing it back with updated ID Engine
//#define FRILLS	0			// Cut out frills for 360K - MIKE MAYNARD

#define ARTSEXTERN


/*
=============================================================================

							MACROS

=============================================================================
*/


#define COLORBORDER(color)		asm{mov	dx,STATUS_REGISTER_1;in al,dx;\
	mov dx,ATR_INDEX;mov al,ATR_OVERSCAN;out dx,al;mov al,color;out	dx,al;\
	mov	al,32;out dx,al};

#define SIGN(x) 	((x)>0?1:-1)
#define ABS(x) 		((int)(x)>0?(x):-(x))
#define LABS(x) 	((long)(x)>0?(x):-(x)

// 	Redefine some functions from Omnispeak as Macros
#define TI_ForeTop(tile)	(tinf[NORTHWALL+tile])
#define TI_ForeRight(tile)	(tinf[EASTWALL+tile])
#define TI_ForeBottom(tile)	(tinf[SOUTHWALL+tile])
#define TI_ForeLeft(tile)	(tinf[WESTWALL+tile])
#define TI_ForeAnimTile(tile)   ((signed char)(tinf[MANIM+tile]))
#define TI_ForeMisc(tile)   	(tinf[INTILE+tile])

#define CA_TileAtPos(x,y,plane) (*((unsigned far *)mapsegs[plane] + mapbwidthtable[y]/2+x))
/*
=============================================================================

						GLOBAL CONSTANTS

=============================================================================
*/

#define	MAXACTORS	MAXSPRITES

#define ACCGRAVITY	3
#define SPDMAXY		80

#define BLOCKSIZE	(8*PIXGLOBAL)		// for positioning starting actors

//
// movement scrolling edges
//
#define SCROLLEAST (TILEGLOBAL*11)
#define SCROLLWEST (TILEGLOBAL*9)
#define SCROLLSOUTH (TILEGLOBAL*8)
#define SCROLLNORTH (TILEGLOBAL*4)

#define CLIPMOVE	24			// speed to push out of a solid wall

#define GAMELEVELS	17

#define PLATFORMBLOCK	31			// InfoTile value 

/*
=============================================================================

							 TYPES

=============================================================================
*/


struct CK_object;

typedef enum CK_Difficulty
{
	D_NotPlaying = 0,
	D_Easy = 1,
	D_Normal = 2,
	D_Hard = 3
} CK_Difficulty;

// see WL_DEF.H
typedef enum CK_Controldir {
        CD_north = 0,
        CD_east = 1,
        CD_south = 2,
        CD_west = 3
} CK_Controldir;

typedef enum CK_Dir {
        Dir_east = 0,
        Dir_northeast = 1,
        Dir_north = 2,
        Dir_northwest = 3,
        Dir_west = 4,
        Dir_southwest = 5,
        Dir_south = 6,
        Dir_southeast = 7,
        Dir_nodir = 8
} CK_Dir;

typedef enum CK_ClassType {
	CT_nothing = 0,
	CT_Friendly = 1,
	CT_Player = 2,
	CT_Stunner = 3,
	CT_EnemyShot = 4,
	CT_Item = 5,
	CT_Platform = 6,
	CT_StunnedCreature = 7,
	CT_MapFlag = 8,
} CK_ClassType;

typedef enum CK_LevelState
{
  LS_Playing = 0,             // In Level
  LS_Died = 1,                // Keen Died
  LS_LevelComplete = 2,       // Level Completed
  LS_CouncilRescued = 3,      // Rescued Council Member (Keen 4)
  LS_AboutToRecordDemo = 4,   // About to Record Demo
  LS_StartingNew = 5,	      // About to start a new game
  LS_StartingSaved = 6,	      // Resuming a saved game
  LS_Aborted	   = 8,
  LS_DestroyedQED = 15,       // Destroyed QED (Keen 5)

  // Unauthentic level states
  LS_WarpToLevel,
} CK_LevelState;

extern CK_LevelState ck_levelState;

// This struct is 0x58 bytes big in Keen5
// It must be preserved if omnispeak savegames are to be compatible 
// with those of vanilla keen
typedef struct CK_GameState
{
	unsigned mapPosX;
	unsigned mapPosY;
	unsigned levelsDone[25];  // Number of levels complete
	long keenScore;			// Keen's score. (This _is_ signed, by the looks of all the 'jl' instructions)
	long nextKeenAt;			// Score keen will get a new life at.
	int numShots;
	int numVitalin;
	int securityCard;
	int word_4729C;
	int fusesRemaining;
	int keyGems[4];
	int currentLevel;
	int numLives;			// Number of lives keen has.
	CK_Difficulty difficulty;		// Difficulty level of current game
	//struct CK_object *platform;  // This was a 16-bit pointer in DOS Keen5.exe

	int generatorsDestroyed;
	boolean jumpCheat;				// Is the jump cheat enabled? (Not in Keen5 gamestate struct)
} CK_GameState;

extern CK_GameState ck_gameState;

typedef enum CK_ActionType
{
	AT_UnscaledOnce,			// Unscaled Motion, Thinks once.
	AT_ScaledOnce,				// Scaled Motion, Thinks once.
	AT_Frame,					// No Motion, Thinks each frame (doesn't advance action)
	AT_UnscaledFrame,			// Unscaled Motion, Thinks each frame
	AT_ScaledFrame				// Scaled Motion, Thinks each frame
} CK_ActionType;

typedef struct CK_action
{
	int chunkLeft;
	int chunkRight;
	CK_ActionType type;
	int protectAnimation, stickToGround; 	// See KeenWiki: Galaxy Action Parameters (lemm/levelass)
	int timer;
	int velX, velY;
	void (*think)(struct CK_object *obj);
	void (*collide)(struct CK_object *obj, struct CK_object *other);
	void (*draw)(struct CK_object *obj);
	struct CK_action *next;
} CK_action;

typedef enum CK_objActive
{
	OBJ_INACTIVE = 0,
	OBJ_ACTIVE = 1,
	OBJ_ALWAYS_ACTIVE = 2,
	OBJ_EXISTS_ONLY_ONSCREEN = 3
} CK_objActive;

typedef enum CK_clipped
{
	CLIP_not = 0,
	CLIP_normal,
	CLIP_simple,
	CLIP_full, 
} CK_ClipType;

typedef struct CK_objPhysData
{
	// Unit cliping box
	int unitX1;
	int unitY1;
	int unitX2;
	int unitY2;
	int unitXmid;


	// Tile clipping box
	int tileX1;
	int tileY1;
	int tileX2;
	int tileY2;
	int tileXmid;
} CK_objPhysData;

// Because this struct was imported from 32 bit omnispeak to 16 bit dos
// need a second struct with unsigned members for holding absolute positions

typedef struct CK_objPhysDataU
{
	// Unit cliping box
	unsigned unitX1;
	unsigned unitY1;
	unsigned unitX2;
	unsigned unitY2;
	unsigned unitXmid;


	// Tile clipping box
	unsigned tileX1;
	unsigned tileY1;
	unsigned tileX2;
	unsigned tileY2;
	unsigned tileXmid;
} CK_objPhysDataU;

typedef struct CK_object
{
	int type;
	CK_objActive active;
	boolean visible;
	CK_ClipType clipped;
	int timeUntillThink;		// Number of frames to delay calling Think function
	unsigned posX;
	unsigned posY;
	int xDirection;
	int yDirection;
	int deltaPosX;
	int deltaPosY;
	int velX;
	int velY;
	int actionTimer;
	CK_action *currentAction;
	int gfxChunk;
	int zLayer;

	CK_objPhysDataU clipRects;

	// In real keen, these are shared, so we're moving them out.
#if 0
	CK_objPhysData oldRects;
	CK_objPhysData deltaRects;
	int nextX;
	int nextY;
#endif

	//TileInfo for surrounding tiles.
	int topTI, bottomTI, leftTI, rightTI;


	//struct RF_SpriteDrawEntry *sde;
	void *sde;

	int user1;
	int user2;
	int user3;
	int user4;

	struct CK_object *next;
	struct CK_object *prev;
} CK_object;

extern CK_objPhysData ck_oldrects;
extern CK_objPhysData ck_deltarects;
extern int ck_nextX;
extern int ck_nextY;

typedef struct CK_keenState
{
	int jumpTimer;
	int poleGrabTime;
	boolean jumpIsPressed;
	boolean jumpWasPressed;
	boolean pogoIsPressed;
	boolean pogoWasPressed;
	boolean shootIsPressed;
	boolean shootWasPressed;

	boolean keenSliding;
	CK_object *platform;
} CK_keenState;

extern CK_keenState ck_keenState;
extern ControlInfo ck_inputFrame;

typedef struct CK_HighScore
{
	char name[58];
	unsigned long score;
	unsigned arg4;
} CK_HighScore;

extern CK_HighScore ck_highScores[8];



/*
=============================================================================

					 CK_MAIN.C DEFINITIONS

=============================================================================
*/

extern	char	str[80],str2[20];
extern	boolean	singlestep,jumpcheat,godmode,tedlevel;
extern	int extravbls; // Imported from Catacomb 3-D
extern	unsigned	tedlevelnum;

void CK_MeasureMultiline(const char *str, unsigned *w, unsigned *h);

void	DebugMemory (void);
void	TestSprites(void);
int		DebugKeys (void);
void	StartupId (void);
void	ShutdownId (void);
void	Quit (char *error);
void	InitGame (void);
void	DemoLoop (void);
void	main (void);

//TODO:		Variable lost and found... put these somwehere
extern		boolean		showscorebox;
extern		CK_Difficulty ck_startingDifficulty;
extern		int load_game_error, ck_startingSavedGame;
extern		boolean		ck_inHighScores;
extern		int		ck_nextMapNumber;
/*
=============================================================================

					  CK_GAME.C DEFINITIONS

=============================================================================
*/

extern		int			fadecount;
extern		CK_LevelState 		levelState;				// Level State (should probably be enum)


// Rename these to names from Catacomb 3-D (fix conflict with SaveGame struct)
boolean	LoadTheGame/*LoadGame*/ (int file);
boolean	SaveTheGame/*SaveGame*/ (int file);


void CK_NewGame(void);
void CK_GameOver(void);
void CK_ResetGame (void);
void CK_SetLoadingTransition(void);
void FadeAndUnhook (void);
void CK_BeginCacheBox (char *title, unsigned numChunks);
void CK_UpdateCacheBox(void);
void CK_FinishCacheBox(void);
void CK_TryAgainMenu(void);
void CK_MapLevelMarkAsDone(void);
void CK_LoadLevel(boolean doCache);
void CK_StartDemoRecord(void);
void CK_EndDemoRecord (void);
void CK_GameLoop(void);



/*
=============================================================================

					  CK_PLAY.C DEFINITIONS

=============================================================================
*/

// invincibility timer
extern	int ck_invincibilityTimer;

// invincibility switch
extern	boolean ck_godMode;

// Switch to toggle Camera following keen.
extern	boolean ck_scrollDisabled;

extern	ControlInfo	ck_inputFrame;

extern	CK_object ck_objArray[];

extern	CK_object *ck_freeObject;
extern	CK_object *ck_lastObject;
extern	int ck_numObjects;

extern	CK_object *ck_keenObj;
extern	CK_object *ck_scoreBoxObj;

// The rectangle within which objects are active.
extern	int ck_activeX0Tile;
extern	int ck_activeY0Tile;
extern	int ck_activeX1Tile;
extern	int ck_activeY1Tile;

void	SpawnScore (void);
void CK_UpdateScoreBox (CK_object *ob);
void ScoreReact (CK_object *ob);

void CK_CountActiveObjects(void);
void DebugMemory (void);
void CK_ItemCheat(void);
void CK_OverlayForegroundTile(int fgTile, int overlayTile);
void CK_WallDebug(void);
void TestSprites(void);
boolean CK_DebugKeys(void);
void CK_CheckKeys(void);
void CK_HandleInput(void);
void CK_SetupObjArray(void);
CK_object *CK_GetNewObj(boolean nonCritical);
void CK_RemoveObj(CK_object *obj);
int CK_ActionThink(CK_object *obj, int time);
void CK_RunAction(CK_object *obj);
void CK_CentreCamera(CK_object *obj);
void CK_MapCamera( CK_object *keen );
void CK_NormalCamera(CK_object *obj);
void StartMusic(unsigned level, boolean repeat);
void StopMusic(void);
void CK_PlayLoop (void);


/*
=============================================================================

					  CK_INTER.C DEFINITIONS

=============================================================================
*/

void CK_HandleDemoKeys(void);
void CK_FizzleFade(void);
void CK_DrawTerminator(void);
void CK_DrawStarWars(void);
void CK_ShowTitleScreen(void);
void CK_PlayDemo(int demoNumber);
void CK_SubmitHighScore(long score, unsigned arg_4);
void CK_DoHighScores(void);

/*
=============================================================================

					  CK_TEXT.C DEFINITIONS

=============================================================================
*/

void HelpScreens (void);

/*
=============================================================================

					  CK_PHYS.C DEFINITIONS

=============================================================================
*/

void CK_ResetClipRects(CK_object *obj);
void CK_SetOldClipRects(CK_object *obj);
void CK_SetDeltaClipRects(CK_object *obj);
void CK_PhysUpdateX(CK_object *obj, int deltaUnitX);
void CK_PhysUpdateY(CK_object *obj, int deltaUnitY);
void CK_PhysKeenClipDown(CK_object *obj);
void CK_PhysKeenClipUp(CK_object *obj);
boolean CK_NotStuckInWall(CK_object *obj);
void CK_PhysClipVert(CK_object *obj);
void CK_PhysClipHorz(CK_object *obj);
void CK_PhysUpdateNormalObj(CK_object *obj);
void CK_PhysFullClipToWalls(CK_object *obj);
void CK_PhysUpdateSimpleObj(CK_object *obj);
void CK_PhysPushX(CK_object *pushee, CK_object *pusher);
void CK_PhysPushY(CK_object *pushee, CK_object *pusher);
void CK_SetAction(CK_object *obj, CK_action *act);
void CK_SetAction2(CK_object *obj, CK_action *act);
boolean CK_ObjectVisible(CK_object *obj);
void CK_PhysGravityHigh(CK_object *obj);
void CK_PhysGravityMid(CK_object *obj);
void CK_PhysGravityLow(CK_object *obj);
void CK_PhysDampHorz(CK_object *obj);
void CK_PhysAccelHorz(CK_object *obj, int accX, int velLimit);
void CK_PhysAccelVert(CK_object *obj, int accX, int velLimit);

/*
=============================================================================

					  CK_KEEN.C DEFINITIONS

=============================================================================
*/

CK_action CK_ACT_keenStanding;
CK_action CK_ACT_keenRidePlatform;
CK_action CK_ACT_keenLookUp1;
CK_action CK_ACT_keenLookUp2;
CK_action CK_ACT_keenLookDown1;
CK_action CK_ACT_keenLookDown2;
CK_action CK_ACT_keenLookDown3;
CK_action CK_ACT_keenLookDown4;
CK_action CK_ACT_keenIdle;
CK_action CK_ACT_keenBored1;
CK_action CK_ACT_keenBored2;
CK_action CK_ACT_keenBored3;
CK_action CK_ACT_keenBored4;
CK_action CK_ACT_keenBored5;
CK_action CK_ACT_keenBored6;
CK_action CK_ACT_keenOpenBook1;
CK_action CK_ACT_keenOpenBook2;
CK_action CK_ACT_keenOpenBook3;
CK_action CK_ACT_keenOpenBook4;
CK_action CK_ACT_keenReadBook1;
CK_action CK_ACT_keenReadBook2;
CK_action CK_ACT_keenReadBook3;
CK_action CK_ACT_keenStowBook1;
CK_action CK_ACT_keenStowBook2;
CK_action CK_ACT_keenStowBook3;
CK_action CK_ACT_keenNot;
CK_action CK_ACT_keenDie0;
CK_action CK_ACT_keenDie1;
CK_action CK_ACT_keenRun1;
CK_action CK_ACT_keenRun2;
CK_action CK_ACT_keenRun3;
CK_action CK_ACT_keenRun4;
CK_action CK_ACT_keenShoot1;
CK_action CK_ACT_keenShoot2;
CK_action CK_ACT_keenShootUp1;
CK_action CK_ACT_keenShootUp2;
CK_action CK_ACT_keenJump1;
CK_action CK_ACT_keenJump2;
CK_action CK_ACT_keenFall1;
CK_action CK_ACT_keenFall2;
CK_action CK_ACT_keenJumpShoot1;
CK_action CK_ACT_keenJumpShoot2;
CK_action CK_ACT_keenJumpShoot3;
CK_action CK_ACT_keenJumpShootUp1;
CK_action CK_ACT_keenJumpShootUp2;
CK_action CK_ACT_keenJumpShootUp3;
CK_action CK_ACT_keenJumpShootDown1;
CK_action CK_ACT_keenJumpShootDown2;
CK_action CK_ACT_keenJumpShootDown3;
CK_action CK_ACT_keenPogo1;
CK_action CK_ACT_keenPogo2;
CK_action CK_ACT_keenPogo3;
CK_action CK_ACT_keenHang1;
CK_action CK_ACT_keenHang2;
CK_action CK_ACT_keenPull1;
CK_action CK_ACT_keenPull2;
CK_action CK_ACT_keenPull3;
CK_action CK_ACT_keenPull4;
CK_action CK_ACT_keenPull5;
CK_action CK_ACT_keenPoleSit;
CK_action CK_ACT_keenPoleUp1;
CK_action CK_ACT_keenPoleUp2;
CK_action CK_ACT_keenPoleUp3;
CK_action CK_ACT_keenPoleDown1;
CK_action CK_ACT_keenPoleDown2;
CK_action CK_ACT_keenPoleDown3;
CK_action CK_ACT_keenPoleDown4;
CK_action CK_ACT_keenPoleShoot1;
CK_action CK_ACT_keenPoleShoot2;
CK_action CK_ACT_keenPoleShootUp1;
CK_action CK_ACT_keenPoleShootUp2;
CK_action CK_ACT_keenPoleShootDown1;
CK_action CK_ACT_keenPoleShootDown2;

CK_action CK_ACT_keenShot1;
CK_action CK_ACT_keenShot2;
CK_action CK_ACT_keenShot3;
CK_action CK_ACT_keenShot4;

CK_action CK_ACT_keenShotHit1;
CK_action CK_ACT_keenShotHit2;

CK_action CK_ACT_keenSlide;

CK_action CK_ACT_keenPlaceGem;

CK_action CK_ACT_keenPressSwitch1;
CK_action CK_ACT_keenPressSwitch2;

CK_action CK_ACT_keenEnterDoor1;
CK_action CK_ACT_keenEnterDoor2;
CK_action CK_ACT_keenEnterDoor3;
CK_action CK_ACT_keenEnterDoor4;
CK_action CK_ACT_keenEnterDoor5;
CK_action CK_ACT_keenEnterDoor6;
CK_action CK_ACT_keenEnteredDoor;


void CK_KeenColFunc(CK_object *a, CK_object *b);
void CK_IncreaseScore(int score);
void CK_SpawnKeen(int tileX, int tileY, int direction);
void CK_KeenGetTileItem(int tileX, int tileY, int itemNumber);
void CK_GetVitalin(int tileX, int tileY);
void CK_KeenGetTileVitalin(int tileX, int tileY);
void CK_KeenCheckSpecialTileInfo(CK_object *obj);
void CK_KeenPressSwitchThink(CK_object *obj);
boolean CK_KeenPressUp(CK_object *obj);
void CK_KeenSlide(CK_object *obj);
void CK_KeenEnterDoor0(CK_object *obj);
void CK_KeenEnterDoor1(CK_object *obj);
void CK_KeenEnterDoor(CK_object *obj);
void CK_KeenPlaceGem(CK_object *obj);
void CK_KeenRidePlatform(CK_object *obj);
boolean CK_KeenTryClimbPole(CK_object *obj);
void CK_KeenRunningThink(CK_object *obj);
void CK_HandleInputOnGround(CK_object *obj);
void CK_KeenStandingThink(CK_object *obj);
void CK_KeenLookUpThink(CK_object *obj);
void CK_KeenLookDownThink(CK_object *obj);
void CK_KeenDrawFunc(CK_object *obj);
void CK_KeenRunDrawFunc(CK_object *obj);
void CK_KeenReadThink(CK_object *obj);
void CK_KeenJumpThink(CK_object *obj);
void CK_KeenJumpDrawFunc(CK_object *obj);
void CK_KeenPogoBounceThink(CK_object *obj);
void CK_KeenPogoThink(CK_object *obj);
void CK_KeenBreakFuse(int x, int y);
void CK_KeenPogoDrawFunc(CK_object *obj);
void CK_KeenSpecialColFunc(CK_object *obj, CK_object *other);
void CK_KeenSpecialDrawFunc(CK_object *obj);
void CK_KeenHangThink(CK_object *obj);
void CK_KeenPullThink1(CK_object *obj);
void CK_KeenPullThink2(CK_object *obj);
void CK_KeenPullThink3(CK_object *obj);
void CK_KeenPullThink4(CK_object *obj);
void CK_KeenDeathThink(CK_object *obj);
void CK_KillKeen(void);
void CK_KeenPoleHandleInput(CK_object *obj);
void CK_KeenPoleSitThink(CK_object *obj);
void CK_KeenPoleUpThink(CK_object *obj);
void CK_KeenPoleDownThink(CK_object *obj);
void CK_KeenPoleDownDrawFunc(CK_object *obj);
void CK_SpawnShot(int x, int y, int direction);
void CK_ShotHit(CK_object *obj);
void CK_ShotThink(CK_object *shot);
void CK_ShotDrawFunc(CK_object *obj);
void CK_KeenSpawnShot(CK_object *obj);
void CK_KeenFall(CK_object *obj);
/*
=============================================================================

					  CK_OBJ.C DEFINITIONS

=============================================================================
*/

extern CK_action CK_ACT_DoorOpen1;
extern CK_action CK_ACT_DoorOpen2;
extern CK_action CK_ACT_DoorOpen3;

extern CK_action CK_act_item;
extern CK_action CK_act_item2;

extern CK_action CK_ACT_VitalinNotify1;
extern CK_action CK_ACT_VitalinNotify2;
extern CK_action CK_ACT_VitalinNotify3;
extern CK_action CK_ACT_VitalinNotify4;

extern CK_action CK_ACT_FallPlat0;
extern CK_action CK_ACT_FallPlat1;
extern CK_action CK_ACT_FallPlat2;

void CK_Fall(CK_object *obj);
void CK_Fall2(CK_object *obj);
void CK_Glide(CK_object *obj);
void CK_DeadlyCol(CK_object *o1, CK_object *o2);
void CK_DeadlyCol2(CK_object *o1, CK_object *o2);
void CK_BasicDrawFunc1(CK_object *obj);
void CK_BasicDrawFunc2(CK_object *obj);
void CK_BasicDrawFunc4(CK_object *obj);
void CK_StunCreature(CK_object *creature, CK_object *stunner, CK_action *new_creature_act);
void CK_SpawnItem(int tileX, int tileY, int itemNumber);
void CK_PointItem(CK_object *obj);
void CK_SpawnAxisPlatform(int tileX, int tileY, int direction);
void CK_AxisPlatform(CK_object *obj);
void CK_SpawnFallPlat(int tileX, int tileY);
void CK_FallPlatSit (CK_object *obj);
void CK_FallPlatFall (CK_object *obj);
void CK_FallPlatRise (CK_object *obj);

void CK_DoorOpen(CK_object *obj);
void CK_SecurityDoorOpen(CK_object *obj);




/*
=============================================================================

					  CK_MAP.C DEFINITIONS

=============================================================================
*/

extern CK_action CK_ACT_MapKeenStart;
extern CK_action CK_ACT_MapKeenWave0;
extern CK_action CK_ACT_MapKeenWave1;
extern CK_action CK_ACT_MapKeenWave2;
extern CK_action CK_ACT_MapKeenWave3;
extern CK_action CK_ACT_MapKeenWave4;
extern CK_action CK_ACT_MapKeenWalk0;
extern CK_action CK_ACT_MapFlag0;
extern CK_action CK_ACT_MapFlag1;
extern CK_action CK_ACT_MapFlag2;
extern CK_action CK_ACT_MapFlag3;

void CK_SpawnMapKeen(int tileX, int tileY);
void CK_ScanForLevelEntry(CK_object * obj);
void CK_MapKeenStill(CK_object * obj);
void CK_MapKeenWalk(CK_object * obj);
void CK_MapMiscFlagsCheck(CK_object *keen);
void CK_MapFlagSpawn(unsigned tileX, unsigned tileY);
void CK_SpawnFlippingFlag(unsigned tileX, unsigned tileY);
void CK_KeenFlagThrown (CK_object *obj);
void CK_KeenFlagSearch (CK_object *obj);
void CK_KeenFlagLands(CK_object *obj);
/*
=============================================================================

					  CK_MISC.C DEFINITIONS

=============================================================================
*/
extern const char *ck_levelNames[];
void CK_ScanInfoLayer(void);

