/* Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 * Copyright (C) 1993-2014 Flat Rock Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

// KD_GAME.C

#include "CK_DEF.H"
#pragma	hdrstop

/*
=============================================================================

						 LOCAL CONSTANTS

=============================================================================
*/

#define RLETAG  0xABCD

/*
=============================================================================

						 GLOBAL VARIABLES

=============================================================================
*/

int			fadecount;

/*
=============================================================================

						 LOCAL VARIABLES

=============================================================================
*/

int			ck_lastLevelFinished;
//==========================================================================



//===========================================================================

void CK_NewGame(void)
{
	// TODO: Zero the ck_gameState
	memset(&ck_gameState, 0, sizeof(ck_gameState));
	ck_gameState.nextKeenAt = 20000;
	ck_gameState.numLives = 3;
	ck_gameState.numShots = 5;
	ck_gameState.currentLevel = 0;
}

void CK_GameOver(void)
{
	// HACK: US_CPrint modifies strings. So don't use
	// literals directly. (char * is NOT ok; Use char arrays.)
	static char gameOverString[] = "Game Over!";

	US_CenterWindow(16, 3);
	US_CPrint(gameOverString);
	VW_UpdateScreen(); //VL_Present();
	//TODO: Wait 4*70 tics
	IN_WaitForKey();
}
boolean
SaveTheGame(int handle)
{
	int i;
	unsigned bufsize;
	memptr compressed;
	CK_object *obj;


	ck_keenState.platform = NULL;
	if (!CA_FarWrite(handle, (byte far *)&ck_gameState, sizeof (ck_gameState)))
		return false;

	bufsize = mapwidth * mapheight * 2;
	MM_GetPtr(&compressed, bufsize);

	// Save three compressed map planes
	for (i = 0; i < 3; i++)
	{
		unsigned planesize = CA_RLEWCompress((unsigned huge *)mapsegs[i], bufsize, (unsigned huge *)compressed + 1, RLETAG);
		*(unsigned huge *)(compressed) = planesize;

		if (!CA_FarWrite(handle, (byte far *)compressed, planesize+2))
		{
			MM_FreePtr(&compressed);
			return false;
		}
	}

	// Save the object list
	for (obj = ck_keenObj; obj; obj = obj->next)
	{
		if (!CA_FarWrite(handle, (byte far *)obj, sizeof(CK_object)))
		{
			MM_FreePtr(&compressed);
			return false;
		}
	}


	MM_FreePtr(&compressed);
	return true;
}

// TODO: Level Names here

boolean
LoadTheGame(int handle)
{
	int i; 
	unsigned bufsize, cmplen;
	memptr buf;
	CK_object *objprev, *objnext, *more_obj, *new_object;

	if( !CA_FarRead( handle, (byte far *)&ck_gameState, sizeof (ck_gameState) ) )
		return false;

	ca_levelbit >>= 1;
	ca_levelnum--;
	CK_LoadLevel(false);
	if( mmerror ) {
		mmerror = 0;
		US_CenterWindow( 20, 8 );
		PrintY += 20;
		US_CPrint( "Not enough memory\nto load game!" );
		VW_UpdateScreen();
		IN_Ack();//wait_for_button();
		return false;
	}

	ca_levelbit <<= 1;
	ca_levelnum++;
	bufsize = mapwidth * mapheight * 2;
	MM_BombOnError( true );
	MM_GetPtr(&buf, bufsize );
	MM_BombOnError( false );
	if( mmerror ) {
		mmerror = 0;
		US_CenterWindow( 20, 8 );
		PrintY += 20;
		US_CPrint( "Not enough memory\nto load game!" );
		VW_UpdateScreen();
		IN_Ack(); //wait_for_button();
		return false;
	}


	/* Read the level data from the file */
	for( i = 0; i < 3; i++ ) {
		if( !CA_FarRead( handle, (byte far *)&cmplen, 2 ) ) {
			MM_FreePtr( &buf );
			return false;
		}
		if( !CA_FarRead( handle, (byte far *)buf, cmplen ) ) {
			MM_FreePtr( &buf );
			return 0;
		}
		CA_RLEWexpand( (unsigned huge *)buf, (unsigned huge *)mapsegs[i], bufsize, RLETAG );
	}
	MM_FreePtr( &buf );

	CK_SetupObjArray();
	new_object = ck_keenObj;

	/* Read the first object (keen) from the list */
	objprev = new_object->prev;
	objnext = new_object->next;
	if( !CA_FarRead( handle, (byte far *)new_object, sizeof(CK_object) ) )
		return 0;
	new_object->prev = objprev;
	new_object->next = objnext;

	new_object->visible = 1;
	new_object->sde = NULL;
	new_object = ck_scoreBoxObj;

	while( 1 ) {
		/* Read the object from the file and put it in the list */
		objprev = new_object->prev;
		objnext = new_object->next;
		if( !CA_FarRead( handle, (byte far *)new_object, sizeof(CK_object) ) )
			return 0;
		more_obj = new_object->next;
		new_object->prev = objprev;
		new_object->next = objnext;

		new_object->visible = 1;
		new_object->sde = 0;

		if( new_object->type == CT_StunnedCreature ) {
			new_object->user3 = NULL;
		} else if( new_object->type == CT_Platform ) {
			new_object->user2 = new_object->user3 = 0;
		}

		/* If this is the last object in the saved list, exit the loop */
		if (!more_obj)
			break;

		/* Otherwise we add a new object */
		new_object = CK_GetNewObj( false );
	}

	ck_scoreBoxObj->user1 = -1;
	ck_scoreBoxObj->user2 = -1;
	ck_scoreBoxObj->user3 = -1;
	ck_scoreBoxObj->user4 = -1;

	return true;

}

void CK_ResetGame (void)
{
	CK_NewGame();
	ca_levelnum--;
	ca_levelbit >>= 1;
	CA_ClearMarks();
	ca_levelbit <<= 1;
	ca_levelnum++;

}

//===========================================================================


//===========================================================================
//
// LEVEL CACHING 
//
//===========================================================================

void CK_SetLoadingTransition(void)
{
	VW_FadeOut();
	fadecount = 0;
	RF_SetRefreshHook (&FadeAndUnhook);
}
/*
==========================
=
= FadeAndUnhook
=
= Latch this onto the refresh so the screen only gets faded in after two
= refreshes.  This lets all actors draw themselves to both pages before
= fading the screen in.
=
==========================
*/

// OMNISPEAK: UpdateFadeDrawing
void FadeAndUnhook (void)
{
	if (++fadecount==2)
	{
		VW_FadeIn ();
		RF_SetRefreshHook (NULL);
		lasttimecount = TimeCount;	// don't adaptively time the fade
	}
}

// Cache Box Routines
// These are accessed as callbacks by the caching manager

static int ck_cacheCountdownNum, ck_cacheBoxChunksPerPic, ck_cacheBoxChunkCounter;

extern const char *ck_levelNames[];
extern char *ck_levelEntryTexts[]; 

// The string may be temporarily modified by US_CPrint, hence it's non-const.
void CK_BeginCacheBox (char *title, unsigned numChunks)
{
	int i;
	unsigned w, h;

	// Vanilla keen checks if > 2kb free here
	// If not, it doesn't cache all of the keen counting down graphics
	// But not necessary for omnispeak
	if (MM_TotalFree() > 2048)
	{
		ck_cacheCountdownNum = 0;
	}
	else
	{
		ck_cacheCountdownNum = 5;
	}

	ck_cacheCountdownNum = 0;

	// Cache the Keen countdown graphics
	for (i = 0; i < 6; i++)
	{
		// TODO: Episode independence
		CA_CacheGrChunk(CD_KEENPIC + i);
		grneeded[CD_KEENPIC + i] &= ~ca_levelbit;

		// If a pic can't be cached, forget updating the hand pics
		// by setting the countdown counter at 5
		if (!grsegs[CD_KEENPIC + i])
		{
			// mmerror = 0;
			ck_cacheCountdownNum = 5;
			break;
		}

		MM_SetPurge(grsegs + CD_KEENPIC + i, 3);
	}

	US_CenterWindow(26, 8);

	if (grsegs[CD_KEENPIC])
		VWB_DrawPic(WindowX, WindowY, CD_KEENPIC);
	else
		ck_cacheCountdownNum = 5;

	grneeded[CD_KEENPIC] &= !ca_levelbit;
	WindowW -= 48;
	WindowX += 48;
	// Omnispeak FIXME: Start printX at the right spot
	// The following line is not present in Keen 5
	PrintX = WindowX;
	CK_MeasureMultiline(title, &w, &h);
	PrintY += (WindowH - h) / 2 - 4;
	US_CPrint(title);
	VW_UpdateScreen(); //VL_Present();

	ck_cacheBoxChunkCounter = ck_cacheBoxChunksPerPic = numChunks / 6;

	if (!ck_cacheBoxChunksPerPic && !ck_cacheCountdownNum)
	{
		ck_cacheCountdownNum = 5;
		if (grsegs[CD_ONEPIC])
		{

			VWB_DrawPic(WindowX - 24, WindowY + 40, CD_ONEPIC);
		}

		VW_UpdateScreen();
	}
}

void CK_UpdateCacheBox()
{
	ck_cacheBoxChunkCounter--;

	if ( ck_cacheBoxChunkCounter == 0 && ck_cacheCountdownNum <= 4 )
	{

		ck_cacheBoxChunkCounter = ck_cacheBoxChunksPerPic;
		if ( grsegs[CD_FIVEPIC + ck_cacheCountdownNum] )
		{
			VWB_DrawPic( WindowX - 24, WindowY + 40, CD_FIVEPIC + ck_cacheCountdownNum);
			VW_UpdateScreen();
		}
		ck_cacheCountdownNum++;
	}
}

void CK_FinishCacheBox()
{

}

void CK_TryAgainMenu()
{
	unsigned w, h; 
	int y1, y2, sel;
	int y, clr;

	char buf[80];

	// HACK: US_CPrint modifies strings. So don't use
	// literals directly. (char * is NOT ok; Use char arrays.)
	static char youDidntMakeItPastString[] = "You didn't make it past";
	static char tryAgainString[] = "Try Again";
	static char exitToArmageddonString[] = "Exit to Krisle";

	/* Copy and measure the level name */
	strcpy( buf, ck_levelNames[mapon] );
	CK_MeasureMultiline( buf, &w, &h );

	/* Take away all gems */
	memset( ck_gameState.keyGems, 0, sizeof (ck_gameState.keyGems) );

	/* If lives remain, see if they want to try this level again */
	if ( --ck_gameState.numLives >= 0 )
	{
		//VW_FixRefreshBuffer();
		US_CenterWindow( 20, 8 );
		PrintY+= 3;
		US_CPrint(youDidntMakeItPastString);
		y1 = PrintY + 22;

		/* Center the level name vertically */
		if ( h < 15 )
			PrintY += 4;	
		US_CPrint(buf);

		PrintY = y1 + 2;
		US_CPrint(tryAgainString);
		PrintY += 4;
		y2 = PrintY - 2;
		US_CPrint(exitToArmageddonString);

		IN_ClearKeysDown();
		sel = 0;
		while ( 1 )
		{
			VW_WaitVBL(1);

			/* Decide which selection to draw */
			if ( sel != 0 )
				y = y2;
			else
				y = y1;

			/* Choose a color to draw it in */
			if ( (TimeCount >> 4) & 1 )
				clr = 12;
			else
				clr = 1;

			/* And draw the selection box */
			VWB_Hlin( WindowX + 4, WindowX + WindowW - 4, y, clr );
			VWB_Hlin( WindowX + 4, WindowX + WindowW - 4, y + 1, clr );
			VWB_Hlin( WindowX + 4, WindowX + WindowW - 4, y + 12, clr );
			VWB_Hlin( WindowX + 4, WindowX + WindowW - 4, y + 13, clr );
			VWB_Vlin( y + 1, y + 11, WindowX + 4, clr );
			VWB_Vlin( y + 1, y + 11, WindowX + 5, clr );
			VWB_Vlin( y + 1, y + 11, WindowX + WindowW - 4, clr );
			VWB_Vlin( y + 1, y + 11, WindowX + WindowW - 5, clr );
			// VL_Present();
			VW_UpdateScreen();

			/* Erase the box for next time */
			VWB_Hlin( WindowX + 4, WindowX + WindowW - 4, y, 15 );
			VWB_Hlin( WindowX + 4, WindowX + WindowW - 4, y + 1, 15 );
			VWB_Hlin( WindowX + 4, WindowX + WindowW - 4, y + 12, 15 );
			VWB_Hlin( WindowX + 4, WindowX + WindowW - 4, y + 13, 15 );
			VWB_Vlin( y + 1, y + 11, WindowX + 4, 15 );
			VWB_Vlin( y + 1, y + 11, WindowX + 5, 15 );
			VWB_Vlin( y + 1, y + 11, WindowX + WindowW - 4, 15 );
			VWB_Vlin( y + 1, y + 11, WindowX + WindowW - 5, 15 );

			/* If they press Esc, they want to go back to the Map */
			if ( LastScan == sc_Escape )
			{
				ck_gameState.currentLevel = 0;
				IN_ClearKeysDown();
				return;
			}

			IN_ReadControl( 0, &ck_inputFrame );
			if ( ck_inputFrame.button0 || ck_inputFrame.button1 || LastScan == sc_Enter || LastScan == sc_Space )
			{
				/* If they want to go back to the Map, set the current level to zero */
				if ( sel != 0 )
					ck_gameState.currentLevel = 0;
				return;
			}

			if ( ck_inputFrame.yaxis == -1 || LastScan == sc_UpArrow )
			{
				sel = 0;
			}
			else if ( ck_inputFrame.yaxis == 1 || LastScan == sc_DownArrow )
			{
				sel = 1;
			}
		} /* while */
	}
}


void CK_MapLevelMarkAsDone(void)
{
	int y, x, level, i;
	unsigned flags, w;
	unsigned far *pw;

	i = 0;
	pw = mapsegs[2];

	/* Look through the map for level-related tiles */
	for ( y = 0; y < mapheight; y++ )
	{
		for ( x = 0; x < mapwidth; x++, pw++, i++ )
		{
			w = *pw;
			level = w & 0xFF;
			if ( level >= 1 && level <= 24 && ck_gameState.levelsDone[level] )
			{	/* Is this a level tile */
				flags = w >> 8;
				/* Set the info tile at this position to 0 */
				*pw = 0;
				if ( flags == 0xD0 )
				{	/* If this is a 'blocking' part of the level */
					/* Set the foreground tile at this position to 0 also (remove the fences) */
					*(mapsegs[1] + y * mapwidth + x) = 0;
				}
				else if ( flags == 0xF0 )
				{	/* If this is the flag holder for the level */
					if ( ck_lastLevelFinished == level )
						CK_SpawnFlippingFlag(x,y);
					else
						CK_MapFlagSpawn(x,y);
				}
			}
		}
	}

}

/*
==========================
=
= SetupGameLevel
=
= Load in map mapon and cache everything needed for it
=
==========================
*/

extern const int level_music[];
void CK_LoadLevel(boolean doCache)
{
	int i;

	if (DemoMode != demo_Off)
	{
		// If we're recording or playing back a demo, the game needs
		// to be deterministic. Seed the RNG at 0 and set difficulty
		// to normal.
		US_InitRndT(false);
		ck_gameState.difficulty = D_Normal;
	}
	else
	{
		US_InitRndT(true);
	}

	CA_CacheMap(ck_gameState.currentLevel);
	RF_NewMap();
	CA_ClearMarks();

	CK_SetupObjArray();
	
	CK_ScanInfoLayer();

	if (mapon == 0)
	{
		CK_MapLevelMarkAsDone(); 
	}

	RF_MarkTileGraphics();
	MM_BombOnError(false);
	CA_LoadAllSounds();

	// Cache Marked graphics and draw loading box
	if (doCache)
	{
		static char demoString[] = "DEMO";
		

		// HACK: US_CPrint modifies strings. So don't use
		// literals directly. (char * is NOT ok; Use char arrays.)
		if (ck_inHighScores)
		{
			CA_CacheMarks(NULL);
		}
		else if (DemoMode != demo_Off)
		{
			CA_CacheMarks(demoString);
		}
		else
		{
			CA_CacheMarks(ck_levelEntryTexts[mapon]);
		}
	}

	// Fade the screen
	MM_BombOnError(true);

	if (mmerror || !doCache)
		return;

	CK_SetLoadingTransition();

}




//===========================================================================
//
// MENUS
//
// TODO: CacheBox
// Begin/Update/Finish CacheBox
// TODO: CK_TryAgain Menu
//===========================================================================

//===========================================================================
//
// DEMO PLAYBACK
// 
//===========================================================================


void CK_StartDemoRecord(void)
{
	VW_FixRefreshBuffer();
	US_CenterWindow(30, 3);
	PrintY += 6;
	US_Print("Record a demo from level(0-2)");
	VW_UpdateScreen();

	if (US_LineInput(px, py, str, 0, true, 2, 0))
	{
		int level = atoi(str);

		if (level > 0 && level < NUMMAPS)
		{
			ck_gameState.currentLevel = level;
			ck_levelState = LS_AboutToRecordDemo;
			IN_StartDemoRecord(0x1000);
		}
	};


}

void CK_EndDemoRecord (void)
{

	char path[] = "DEMO?."EXTENSION;
	IN_StopDemo();
	VW_FixRefreshBuffer();
	US_CenterWindow(22, 3);
	PrintY += 6;
	US_Print (" Save as demo #(0-9):");
	VW_UpdateScreen();

	if (US_LineInput(px, py, str, 0, true, 2, 0))
	{
		if (str[0] >= '0' && str[0] <= '9')
		{
			int handle;
			path[4] = str[0];
			if ((handle = open(path, O_BINARY | O_CREAT | O_WRONLY, S_IFREG | S_IREAD | S_IWRITE)) == -1)
			{
				Quit("EndDemoRecord: Cannot write demo file!");
			}

			write(handle, &mapon, sizeof(mapon));
			write(handle, &DemoOffset, sizeof(DemoOffset));
			CA_FarWrite(handle, DemoBuffer, DemoOffset);
			close(handle);
		}
	}

	IN_StopDemo();

}

/*
============================
=
= GameLoop
=
= A game has just started (after the cinematic or load game)
=
============================
*/
void CK_GameLoop()
{
	do
	{
		if (ck_levelState != 6)
		{
resetDifficultyAndLevel:
			ck_gameState.difficulty = restartgame;
			restartgame = D_NotPlaying;
loadLevel:
			CK_LoadLevel(true);

			//TODO: If this didn't succeed, return to level 0.
		}

replayLevel:
		ck_scrollDisabled = false;
		SD_WaitSoundDone();
		CK_PlayLoop();

		if (ck_levelState != 6)
		{
			memset(ck_gameState.keyGems, 0, sizeof (ck_gameState.keyGems));
			// TODO: This is probably (but not necessarily) Keen 5 specific
			ck_gameState.securityCard = 0;
		}

		//TODO: Some TED launching stuff


		ck_lastLevelFinished = -1;
		switch (ck_levelState)
		{
		case LS_Died:
			CK_TryAgainMenu();
			//ck_gameState.currentLevel = ck_nextMapNumber;
			break;

		case LS_LevelComplete:
		case 7:
		case 13:
			if (mapon == 0)
			{
				unsigned bufsave = bufferofs;
				bufferofs = displayofs;

				US_CenterWindow(0x1A, 8);
				PrintY += 0x19;
				US_CPrint("One Moment");

				bufferofs = bufsave;

			}
			else
			{
				//We've beat the level, return to main map.
				//TODO: Mark level as done (and more)
				SD_PlaySound(LEVELEXITSND);
backToMap:
				ck_gameState.currentLevel = 0;
				ck_lastLevelFinished = mapon;
				ck_gameState.levelsDone[mapon] = 1;
				// TODO: If Keen launched with /Demo switch
				// Then function returns here based on mapon
			}
			break;

		// Starting New game
		case 5:
			goto resetDifficultyAndLevel;

		// Loading Game
		case 6:
			goto replayLevel;

		case 8:
			// Quit to Dos
			IN_ClearKeysDown();
			return;

#if 0
		case 14:
			// The level has been ended by fuse destruction
			SD_PlaySound(LEVELEXITSND);
			//ck_lastLevelFinished = mapon;
			ck_gameState.levelsDone[mapon] = 14;
			CK5_FuseMessage();
			ck_gameState.currentLevel = 0;
			break;
#endif

		case 15:
			// The QED was destroyed
			/*
			 * purge_chunks()
			 * RF_Reset();
			 * VW_FixRefreshBuffer();
			 */
			// help_endgame();
#if 0
			CK_SubmitHighScore(ck_gameState.keenScore, 0);
#endif
			return;
			// Warping level
			// This code is added for omnispeak so that the warp functionality works
			// Case 4 normally switches to default
			// UPDATE (Mar 6 2014): Not needed anymore.
		case LS_WarpToLevel:
			// ck_gameState.currentLevel = ck_nextMapNumber;
			break;


		case 3:
		case 9:
		case 10:
		case 11:
		case 12:
			break;
		}

		if (ck_gameState.numLives < 0)
			break;

		goto loadLevel; //livesLeft >= 0
	}	while (true);

	// Keen4/6 simple game over
	CK_GameOver();

	CK_SubmitHighScore(ck_gameState.keenScore, 0);
}

