/* Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 * Copyright (C) 1993-2014 Flat Rock Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

// KD_INTER.C

#include "CK_DEF.H"
#include <stddef.h>
#pragma	hdrstop

/*
=============================================================================

						 LOCAL CONSTANTS

=============================================================================
*/

/*
=============================================================================

						 GLOBAL VARIABLES

=============================================================================
*/

void CK_PlayDemo(int demoNumber);
void CK_SubmitHighScore(long score, unsigned arg_4);
void CK_DoHighScores(void);
void CK_DrawTerminator(void);
void CK_DrawStarWars(void);


/*
=============================================================================

						 LOCAL VARIABLES

=============================================================================
*/

/*
 * CK_INTER: Holds an assortment of screen drawing and state switching routines
 */


boolean ck_inHighScores = false;

void CK_HandleDemoKeys()
{
	if (!LastScan)
		return;

	/*
	 * if Keen started with demo parameter
	 * ck_levelState = 5;
	 * startingDifficulty = 2;
	 * IN_ClearKeysDown();
	 * CK_NewGame(); (init_keen_stats)
	 * return;
	 */

	if (LastScan == sc_F1)
	{
		HelpScreens();
		return;
	}

	// Otherwise, start the wristwatch menu
	US_ControlPanel();
	if (restartgame)
	{
		ck_levelState = 5;
		return;
	}

	if (!loadedgame)
		return;

	ck_levelState = 6;
}

//===========================================================================
//
// TERMINATOR SCREEN
// 
//===========================================================================

#define TERMINATORSCREENWIDTH 248


// Pointers to the two monochrome bitmaps that are scrolled during the intro

typedef struct introbmptypestruct {
	int height, width;
	unsigned linestarts[200];
	byte data[];
} introbmptype;



introbmptype _seg *ck_introKeen;
introbmptype _seg *ck_introCommander;

// segment pointers to various graphic chunks
memptr introbuffer;
memptr introbuffer2;
memptr shiftedCmdrBMPsegs[8];

byte terminator_blackmasks[] = {0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01}; 
byte terminator_lightmasks[] = {0x00, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE}; 
byte terminator_palette1[17] = {0,4,4,0x1c,1,1,1,1,0x11,0x11,0x11,0x11,0x13,0x13,0x13,4,0}; 
byte terminator_palette2[17] = {0,1,1,0x11,1,1,1,1,0x04,0x04,0x04,0x04,0x14,0x14,0x14,0x14,0}; 

// chunk #'s for the vertically scrolling pics
int terminator_pics[4] = {I_IDMODPIC, I_ARRANGEDPIC, I_HELPFROMPIC, I_CATFACEPIC};
// storage for the vertically scrolling pics
memptr terminator_pic_memptrs[4];


unsigned ck_introCommanderWidth;
unsigned ck_introScreenWidth;


// type-identified variables
int  ck_currentTermPicTimeRemaining;
int ck_currentTermPicStartTime;

// Which terminator credit image we're drawing
int ck_currentTerminatorCredit;

// What action the terminator credit is currently doing (loading/flying up/pausing/flying out)
int ck_termCreditStage;
// int rowsToDraw;

unsigned *word_499D1;
memptr ck_currentTermPicSeg;

unsigned word_499CB;
// unknown vairables


word word_46CA2[2];
word terminatorPageOn; // always 0 or 1? Maybe a pageflip thing?
word word_46CB0;


word word_499C7;
word terminatorOfs;	 	// Start of the screen buffer; the KEEN graphic is aligned to the left edge of this
byte terminatorPlaneOn;
word ck_currentTermPicNextLineDist;
word ck_currentTermPicSize;
word ck_currentTermPicHeight;
word ck_currentTermPicHalfWidth;
word ck_currentTermPicWidth;
word ck_termPicHeights[5];
word ck_termPicWidths[5];

word word_499FF;



void DoubleSizeTerminatorPics(int pic);
int AdvanceTerminatorCredit(int elapsedTime);
void ScrollTerminatorCredits(word elapsedTime, word xpixel);
void AnimateTerminator(void);
void TerminatorExpandRLE(word far *src, byte far *dest);
void JoinTerminatorPics(void);
void ZoomOutTerminator_1(word far *arg0, word arg4, int leftOffset, word far *arg8);
void ZoomOutTerminator(void);
void FinishPage(void);

// Caches vertically scrolling terminator pics
// and unsignedly extends each byte of the pic to a word

// Presumably this is because we're drawing in a different graphics mode?
void DoubleSizeTerminatorPics(int pic)
{

	byte far *picptr;
	word far *double_picptr;
	int width, height;
	unsigned i, size;
	int picchunk, pictableindex;

	// Cache the standard-sized picture
	picchunk = terminator_pics[pic];
	CA_CacheGrChunk(picchunk);

	// Allocate memory for the expanded picture
	// and save its dimensions
	pictableindex = picchunk - STARTPICS;
	ck_termPicWidths[pic] = width = (pictable + (pictableindex))->width;
	ck_termPicHeights[pic] = height = (pictable + (pictableindex))->height;
	size = width * height * 2;
	MM_GetPtr(&terminator_pic_memptrs[pic], size * 2);

	// Copy the data to the expanded buffer
	picptr = grsegs[picchunk];
	double_picptr = (word far *)terminator_pic_memptrs[pic];

	for (i = 0; i < size; i++)
		*double_picptr++ = *picptr++ * 2;

	// Free the standard-sized picture
	MM_FreePtr(&grsegs[picchunk]); 
}

// Returns the Y-position of the terminator credit
int AdvanceTerminatorCredit(int elapsedTime)
{

	int bx;

	switch(ck_termCreditStage)
	{

	case -1:
		// loading
		ck_currentTermPicSeg = terminator_pic_memptrs[ck_currentTerminatorCredit];
		ck_currentTermPicWidth = ck_termPicWidths[ck_currentTerminatorCredit];
		ck_currentTermPicHalfWidth = (ck_currentTermPicWidth+3)>>1;
		ck_currentTermPicHeight = ck_termPicHeights[ck_currentTerminatorCredit];
		ck_currentTermPicNextLineDist = linewidth - (ck_currentTermPicWidth+1);
		ck_currentTermPicSize = (ck_currentTermPicWidth * ck_currentTermPicHeight) << 1;
		ck_termCreditStage++;

		ck_currentTermPicStartTime = elapsedTime;
		ck_currentTermPicTimeRemaining = 240;
		// no break, flow into case 0

	case 0:

		// Flying up from bottom of screen
		ck_currentTermPicTimeRemaining -= (elapsedTime - ck_currentTermPicStartTime)<<1;

		if (ck_currentTermPicTimeRemaining < 100)
		{
			ck_currentTermPicTimeRemaining = 100;
			ck_termCreditStage++;
		}

		ck_currentTermPicStartTime = elapsedTime;

		return ck_currentTermPicTimeRemaining - (ck_currentTermPicHeight>>1);


	case 1:

		// Pausing in the middle
		if (elapsedTime - ck_currentTermPicStartTime > 200)
		{
			ck_termCreditStage++;
			ck_currentTermPicStartTime = elapsedTime;
		}

		return 100 - (ck_currentTermPicHeight>>1);

	case 2:

		// Flying up and out of the top of the screen
		ck_currentTermPicTimeRemaining -= (elapsedTime - ck_currentTermPicStartTime)<<1;

		if (ck_currentTermPicTimeRemaining < -40)
		{
			ck_currentTermPicTimeRemaining = -40;

			// Are there any more terminator credit images?
			if (++ck_currentTerminatorCredit < 4)
				ck_termCreditStage = -1;
			else
				ck_termCreditStage = 3;
		}

		ck_currentTermPicStartTime = elapsedTime;

		return ck_currentTermPicTimeRemaining - (ck_currentTermPicHeight>>1);
	
	default:
		break;

	}

		return -40;
}

void ScrollTerminatorCredits(word elapsedTime, word xpixel)
{
	word pelpan, var4, varC,var10;

	// Vars are static because BP is used during ASM draw routine
	static int rowsToDraw;


	int creditY1, varA, creditY2, var12, varE;

	pelpan = xpixel & 7;
	var4 = terminatorOfs + (xpixel>>3) + (20 - ck_currentTermPicWidth/2);

	EGAMAPMASK(0xC);

	creditY1 = AdvanceTerminatorCredit(elapsedTime);
	creditY2 = creditY1 + ck_currentTermPicHeight;

	if (creditY2 < 0)
		creditY2 = 0;

	varA = word_46CA2[terminatorPageOn];

	// Erasing the area underneath the credit
	if (creditY2 < 200 && varA > creditY2)
	{
		varC = varA - creditY2;
		var10 = ylookup[creditY2] + var4;

asm		mov es, screenseg;
asm		mov bx, linewidth;				// bx = next line distance
asm		sub bx, ck_currentTermPicHalfWidth;
asm		sub bx, ck_currentTermPicHalfWidth;
asm		mov di, var10;
asm		mov dx, varC;
asm		mov si, ck_currentTermPicHalfWidth;
asm		xor ax, ax;


nextrow1:
asm		mov cx, si;
asm		rep stosw;
asm		add di, bx;
asm		dec dx;
asm		jnz nextrow1;
	}

	// loc_140AB
	if (creditY2 > 200)
		creditY2 = 200;

	word_46CA2[terminatorPageOn] = creditY2;

	rowsToDraw = ck_currentTermPicHeight;
	varE = 0;

	if (creditY1 < 0)
	{
		// clipped by screen top
		varE = -creditY1 * (ck_currentTermPicWidth<<1);
		rowsToDraw += creditY1;
		creditY1 = 0;
	}
	else if (creditY1 + ck_currentTermPicHeight > 200)
	{
		// clipped by screen bot
		rowsToDraw = 200 - creditY1;
	}


	// loc_14108
	word_499CB = varE + ck_currentTermPicSize;

	if (rowsToDraw > 0)
	{
		word_499D1 = shifttabletable[pelpan];

		word_46CB0 = ylookup[creditY1] + var4;

asm		mov bx, varE;
asm		push bp;
asm		mov bp, word_499D1;
asm		mov es, screenseg;
asm		mov ds, ck_currentTermPicSeg;
asm		mov ah, 4;
asm		mov ss:terminatorPlaneOn, ah;

nextplane:
asm		mov dx, SC_INDEX;
asm		mov al, SC_MAPMASK;
asm		out dx, ax;
asm		mov dx, ss:rowsToDraw;
asm		mov di, ss:word_46CB0;

nextrow:
asm		mov cx, ss:ck_currentTermPicWidth;
asm		xor al, al;

nextX:
asm		mov si, [bx];
asm		add bx, 2;
asm		xor ah, ah;
asm		or ax, [bp+si];
asm		stosb;
asm		mov al, ah;
asm		loop nextX;


asm		stosb;
asm		mov word ptr es:[di], 0;
asm		add di, ss:ck_currentTermPicNextLineDist;
asm		dec dx;
asm		jnz nextrow;


asm		mov bx, ss:word_499CB;
asm		mov ah, ss:terminatorPlaneOn;
asm		shl ah, 1;
asm		mov ss:terminatorPlaneOn, ah;
asm		cmp ah, 0x10;
asm		jnz nextplane;

asm		pop bp;				// Restore stack
asm		mov ax, ss;
asm		mov ds, ax;

				
	}

}


// Does the terminator scrolling 
void AnimateTerminator(void)
{

	int var6;
	long delaytime;

	// At the end of the terminator scrolling, left edge of the commander graphic is
	// this many pixels from the Left edge of the Visible screen
	// (i.e., a negative value means an offset to the left)
	int finalCmdrPos;

	// The COMMANDER graphic scrolls this many pixels to the left
	int totalCmdrScrollDist;

	// How far it's currently scrolled
	int elapsedCmdrScrollDist;
	
	// How many pixels the left edge of the screen is away from the start of the KEEN image
	int xpixel;

	// start of the visible area in the screen buffer
	unsigned screenofs;

	unsigned elapsedTime;

	// The pixel modulo 8 that the left edge of the scrolling CMDR graphic is aligned to,
	// relative to the left edge of the buffer
	int cmdrBMPpelpan;

	// Which of the shifted "COMMANDER" graphics to use
	memptr bmpsrcseg;
	
	// variables of unknown type
	int screenWidthInPx, maxTime, var18, var1A, var1C, var1E, var20, var22;

	EGAWRITEMODE(0);
	EGAREADMAP(0);
	ck_introScreenWidth = ck_introKeen->width + 200;

	EGAREADMAP(1);
	
	finalCmdrPos = 120 - ck_introCommander->width;
	screenWidthInPx = 320;
	totalCmdrScrollDist = finalCmdrPos - screenWidthInPx;
	maxTime = ABS(totalCmdrScrollDist);

	terminatorPageOn = terminatorOfs = 0;

	// delay for a tic before starting
	lasttimecount = TimeCount;
	while (lasttimecount == TimeCount)
		;

	lasttimecount = TimeCount;

	// Update the terminator
	for (elapsedTime = 0; elapsedTime <= maxTime; elapsedTime += tics)
	{

		// Reposition the left edge of the screen in direct proportionality to time elapsed
		xpixel = ((long)ck_introScreenWidth*(long)(maxTime-elapsedTime)) / (long)maxTime;

		// Scroll the Credits graphics 
		ScrollTerminatorCredits(elapsedTime, xpixel);

		elapsedCmdrScrollDist = screenWidthInPx + ((long)totalCmdrScrollDist * (long)elapsedTime)/ (long)maxTime;
		elapsedCmdrScrollDist += xpixel&7;

		var6 = (elapsedCmdrScrollDist + 0x800) / 8 - 0x100;
		cmdrBMPpelpan = ((elapsedCmdrScrollDist + 0x800) & 7);

		bmpsrcseg = shiftedCmdrBMPsegs[cmdrBMPpelpan];
		var18 = 0;

		screenofs = terminatorOfs + xpixel/8;

		if (var6 > 0)
		{
			// The COMMANDER graphic has not started to scroll off the left edge of the screen
			var1A = (var6+1)/2;

			if (var6 & 1)
				screenofs--;

			var1C = 21 - var1A;
			var1E = 0;
		}
		else if (var6 > (int)(320/8 - ck_introCommanderWidth))
		{
			// The COMMANDER graphic is clipped on both left and right edges
			var1A = 0;
			var1C = 21;
			var1E = 0;
			var18 -= var6;
		}
		else
		{
			// The COMMANDER graphic is only clipped by the left side of the screen
			var1A = 0;
			var1C = (ck_introCommanderWidth+var6)/2;
			var1E = 21 - var1C;
			var18 -= var6;
		}

		// loc_1445B:
		var20 = ck_introCommanderWidth - (var1C<<1);
		var22 = TERMINATORSCREENWIDTH - ((var1A + var1C + var1E) << 1);

		EGAMAPMASK(2);


asm		mov di, screenofs;
asm		mov es, screenseg;
asm		mov si, var18;
asm		mov word_499FF, si;
asm		mov ds, bmpsrcseg;

// 		For each row
asm		mov dx, 200;

loopb0:
// Write black pixels to the left of COMMANDER
asm		xor ax, ax;
asm		mov cx, var1A;
asm		rep stosw;

// Write COMMANDER bmp data
asm		mov cx, var1C;
asm		rep movsw;

// Write black pixels to the right of COMMANDER
asm		xor ax, ax;
asm		mov cx, var1E;
asm		rep stosw;


asm		test dx, 1;
asm		jnz odd;

even:
asm		mov si, ss:word_499FF;
asm		jmp ahead;

odd:
asm		add si, var20;
asm		mov ss:word_499FF, si;

ahead:
asm		add di, var22;
asm		dec dx;
asm		jnz loopb0;


asm		mov ax, ss;
asm		mov ds, ax;



		// Flip to the back page while panning the screen to the left 
		VW_SetScreen(xpixel/8 + terminatorOfs,xpixel&7);

		// Set the new back page
		if (terminatorPageOn ^=1)
			terminatorOfs = TERMINATORSCREENWIDTH/2;
		else
			terminatorOfs = 0;


		// Delay
		do {
			delaytime = TimeCount;
		} while ((tics = delaytime - lasttimecount) < 2);

		lasttimecount = delaytime;


		// Stop drawing if key pressed
		if (IN_IsUserInput() && LastScan == sc_F1)
			LastScan = sc_Space;

		if (LastScan)
			return;

	}

	word_499C7 = xpixel/8;

}

// RLE-Expands one line of monochrome terminator BMP data
// Each source line is a sequence of WORD data, terminated with an 0xFFFF flag
// Each line of BMP data is a sequence of word values that encodes alternating runs of black and white pixels,
//  always starting with BLACK pixels

void TerminatorExpandRLE(word far *src, byte far *dest)
{
	// cx = next source word
	// si = number of pixels of one color remaining, in one byte, before transition to the alternate color
	// dl = byte to write at the transition between black and white
	register word si, di, cx;
	register byte dl;

	si = dl = 0;

	while ((cx = *src++) != 0xFFFF)
	{
		// Expand a Black Run of pixels
		if ((si += cx) > 7)
		{
			// write the transition byte
			*dest++ = dl;

			// the leftmost pixels of the transition byte will be black
			dl = 0;

			// write complete bytes
			di = si/8-1;
			while (di--)
				*dest++ = 0;

			// count the leftover pixels on this last byte
			si &= 7;
		}

		// loc_145CC
		// Check if we've hit a stop signal
		if ((cx = *src++) == 0xFFFF)
		{
			*dest++ = dl;
			*dest = 0;
			return;
		}

		// the lowest bits in this byte will be black (zero)
		// so we OR the right most remaining bits so that they are drawn white
		dl |= terminator_blackmasks[si];

		// Expand a Light Run of pixels
		if ((si += cx) > 7)
		{
			// write the transition byte
			*dest++ = dl;

			// the leftmost pixels of the transition byte will be light
			dl = 0xFF;

			// write runs of 8 light pixels
			di = si/8-1;
			while (di--)
				*dest++ = 0xFF;

			// count the leftover light pixels on this last byte
			si &= 7;
		}

		// the lowest bites in this byte will be light ( set)
		// so we AND the rightmost remaining bits so that they are drawn black
		dl &= terminator_lightmasks[si];
	}

	// write the final byte, and one padding byte of black
	*dest++ = dl;
	*dest = 0;
}

// Generate an RLE-compressed bitmap with COMMANDER KEEN separated by some space
// Except... this RLE-compressed bitmap seems to be in a slightly different format
// than those stored in the game archive
//
// Here, alternating runlengths of pixels are not specified by their length, but by their start position
// 
void JoinTerminatorPics(void)
{

	// Points to the data in the stitched bmp
	word far *introBMPDataPtr;

	int i;

	// Allocate memory for the COMMANDER KEEN graphic
	MM_GetPtr(&introbuffer2, 30000);

	// Where we start writing the bitmap data
	introBMPDataPtr = MK_FP(introbuffer2, offsetof(introbmptype, data));

	// for each row
	for (i = 0; i < 200; i++)
	{
		word count, inword;
		word far *linestart;

		// Generate a pointer to this line of data
		((introbmptype far *)introbuffer2)->linestarts[i] = FP_OFF(introBMPDataPtr);

		count = 0;
		EGAREADMAP(1);
		
		linestart = MK_FP(ck_introCommander, ck_introCommander->linestarts[i]); 
		inword = *linestart++;

		do 
		{
			*introBMPDataPtr++ = count;
			count += inword;
			inword = *linestart++;

		} while (inword != 0xFFFF);

		// Add some space between the COMMANDER and the KEEN
		count += 80;

		EGAREADMAP(0);


		linestart = MK_FP(ck_introKeen, ck_introKeen->linestarts[i]);
		linestart++;
		inword = *linestart++;

		do {

			*introBMPDataPtr++ = count;
			count += inword;
			inword = *linestart++;

		} while (inword != 0xFFFF);

		*introBMPDataPtr++ = count;
		*introBMPDataPtr++ = 0xFFFF;

	}
}

//ZoomOutTerminator_1(var16, si+yBottom, leftOffset, introbuffer);
void ZoomOutTerminator_1(word far *arg0, word arg4, int leftOffset, word far *arg8)
{

	word si;


	// X coord where the zoomed bitmap is clipped by the left edge of the screen
	int leftclip;

	int dx, var8;

	byte far *vidPtr; // Pointer into video memory
	byte writebyte;

	unsigned pixelrun;  // The width?
	unsigned varA;

	writebyte = pixelrun = 0;

	var8 = 320-leftOffset;

	vidPtr = MK_FP(0xA000 , (ylookup[arg4] + terminatorOfs + word_499C7));

	if (leftOffset < 0)
	{
		// The zooming bitmap is clipped by the left edge of the screen
		// So find out where we start drawing it
		
		leftclip = -leftOffset;
		writebyte = pixelrun = 0;

		do {

			si = *arg0++;
			dx = *(arg8 + si);
			if ( dx > leftclip)
				goto loc_14852;

			si = *arg0++;
			dx = *(arg8 + si);
			if ( dx > leftclip)
				goto loc_147FB;

		} while (1);
	}
	else
	{
		// loc_147C2:
		// The zooming bitmap is not clipped by left edge of screen
		writebyte = 0;
		pixelrun = leftOffset&7;
		vidPtr += leftOffset/8;
		leftclip = 0;
		arg0++;
		goto loc_147E5;
	}


loc_14852:
	do
	{
		// Writing a run of pixels
		varA = dx - leftclip;
		leftclip = dx;

		// get the run of pixels for the first byte
		writebyte |= terminator_blackmasks[pixelrun];

		if ((pixelrun += varA) > 7)
		{

			int cx;
			// write the first byte
			*vidPtr++ = writebyte; 

			// write bytes of pixels
			writebyte = 0xFF;
			for (cx = pixelrun/8 - 1; cx; cx--)
				*vidPtr++ = 0xFF; // = writebyte;	

			// remaining pixels to write
			pixelrun &= 7;
		}

		if (dx > var8)
			return;

		writebyte &= terminator_lightmasks[pixelrun];

loc_147E5:
		si = *arg0++;
		dx = *(arg8 + si);

loc_147FB:

		// Writing a run of pixels
		varA = dx - leftclip;
		leftclip = dx;
		if ((pixelrun += varA)> 7)
		{

			int cx;
			// write the first byte
			*vidPtr++ = writebyte; 

			// write bytes of pixels
			writebyte = 0;
			for (cx = pixelrun/8 - 1; cx; cx--)
				*vidPtr++ = 0; // = writebyte;	

			// remaining pixels to write
			pixelrun &= 7;

		}

		if (dx > var8)
			return;

		if ((si = *arg0++) != 0xFFFF)
			dx = *(arg8 + si);
		else
			break;
	} while (1);

	// Write black until the end of the screen?
	varA = 320 - leftclip;
	if ((pixelrun += varA) > 7)
	{

		int cx;
		// write the first byte
		*vidPtr++ = writebyte; 

		// write bytes of pixels
		writebyte = 0;
		for (cx = pixelrun/8 - 1; cx; cx--)
		       *vidPtr++ = 0; // = writebyte;	

		// remaining pixels to write
		pixelrun &= 7;
		// return;
	}

	return;

}

// The COMMANDER and KEEN RLE-Encoded bitmaps are first joined together into one big
// RLE-encoded bitmap, which is then scaled and translated to its final position.

// NOTE: I think that this still has errors, which are visible when setting maxTime to a large value
// But... it looks good enough at normal game speed
void ZoomOutTerminator(void)
{
	long startingLeftOffset;
	unsigned yBottom, var18;
	unsigned var20[2];
	unsigned var1C, si;

	// The px offset of the left edge of the bitmap from the left edge of the screen
	// (i.e., negative number means graphic is clipped by left edge of screen
	int leftOffset;

	unsigned elapsedTime, maxTime;
	word far* var16;

	// finalHeight looks like final Height?
	word newTime, var8, finalHeight, varC, varE;


	// Set the palette
asm	mov     ax, ds;
asm	mov	es, ax;
asm	mov	dx, offset terminator_palette1;
asm	mov	ax, 0x1002;
asm 	int	0x10;

	EGAREADMAP(1);

	// The starting (negative) offset of the COMMANDER graphic from the left edge of the screen
	startingLeftOffset = 120 - ck_introCommander->width;

	JoinTerminatorPics();


	MM_GetPtr((memptr *)&introbuffer, 5000);
		
	var8 = 256;
	finalHeight = 33;
	var20[0] = var20[1] = 200;
	elapsedTime = 1;
	maxTime = 30; //3000; // Set to large value to slow down the zoom
	

	// elapsedTime seems to be a timer, maxTime is the max time
	while (elapsedTime <= maxTime)
	{

		if (elapsedTime == maxTime)
		{
			// We're done
			var8 = finalHeight;
			leftOffset = 0;
			yBottom = 4;
		}
		else
		{
			// These casts to long don't actually exist in disasm, but are required for 
			// testing large values of maxtime
			var8 = 256 - (((long)(256 - finalHeight) * (long)elapsedTime) / (long)maxTime);

			leftOffset = (startingLeftOffset * (long)(maxTime-elapsedTime))/ (long)maxTime;

			yBottom = (long)(elapsedTime*4)/(long)maxTime;
		}



		// Generate a table of 2500 multiples of var8 * 256
asm		xor    ax, ax;
asm		xor    dx, dx;
asm		mov    cx, 2500;
asm		mov    bx, var8;
asm		mov    es, introbuffer;
asm		xor    di, di;

loop1:

asm		mov    es:[di], ah;
asm		inc    di;
asm		mov    es:[di], dl;
asm		inc    di;
asm		add    ax, bx;
asm		adc    dx, 0;
asm		loop   loop1;

		if (elapsedTime == maxTime)
			leftOffset = 0;
		else
			leftOffset = ((long)(maxTime-elapsedTime) * startingLeftOffset) / (long)maxTime;


		var18 = *((word far *)introbuffer + 200);
		varC =0 ;
		varE = 0x10000L / var8;
		bufferofs = terminatorOfs + word_499C7;

		if (yBottom > 0)
			VW_Bar(0,0,320,yBottom, 0);


		EGAWRITEMODE(0);
		EGAMAPMASK(0xF);

		// Draw each line to the screen
		for (si = 0; si < var18; si++)
		{
			var16 = MK_FP(introbuffer2, ((introbmptype far*)introbuffer2)->linestarts[varC>>8]);
			ZoomOutTerminator_1(var16, si+yBottom, leftOffset, introbuffer);
			varC += varE;
		}

		bufferofs = terminatorOfs + word_499C7;
		var1C = var18 + yBottom;
		if (var20[terminatorPageOn] > var1C)
		{
			VW_Bar(0, var1C, 320,var20[terminatorPageOn],0);
			var20[terminatorPageOn] = var1C;
		}

		// Flip the page
		VW_SetScreen(terminatorOfs + word_499C7, 0);
		if (terminatorPageOn ^= 1)
			terminatorOfs = TERMINATORSCREENWIDTH/2;
		else
			terminatorOfs = 0;

		// loc_14B4F;
		newTime = TimeCount;
		tics = newTime-lasttimecount;
		lasttimecount = newTime;

		if (tics > 8)
			tics = 8;

		if (elapsedTime == maxTime)
			break;

		elapsedTime += tics;

		if (elapsedTime > maxTime)
			elapsedTime = maxTime;


		if (IN_IsUserInput() && LastScan == sc_F1)
			LastScan = sc_Space;

		if (LastScan)
			// return;// should this be break instead? Want to free intro buffers!
			break; 

	}

	MM_FreePtr(&introbuffer);
	MM_FreePtr(&introbuffer2);
}



// The Fizzlefade routine
// Operates by drawing the title graphic into offscreen video memory, then
// using the hardware to copy the source to the display area, pixel-by-pixel

// The same effect can be achieved in omnispeak by drawing the title screen
// to an new surface, then copying from the surface to the screen
void FinishPage(void)
{
	int i;
	byte bitmasks[] = {1, 2, 4, 8, 0x10, 0x20, 0x40, 0x80};
	int columns1[320], rows1[200];
	
	int pageDifference;
	word varE, var8, row1;


	// Generate a random array from 0 to 320

	// Make array
	for (i = 0; i < 320; i++) 
		columns1[i] = i;

	// Randomly switch pairs of integers
	for (i =0; i< 320; i++)
	{
		int indexToSwitch = (320L * (long)rand()) / (long)(RAND_MAX+1);  
		int valueToSwitch = columns1[indexToSwitch];
		columns1[indexToSwitch] = columns1[i];
		columns1[i] = valueToSwitch;
	}

	for (i=0; i < 200; i++)
		rows1[i] = columns1[i];



	VW_SetDefaultColors();

	if (terminatorPageOn)
	{
		bufferofs = word_499C7 + TERMINATORSCREENWIDTH/2;
		displayofs = word_499C7;
	}
	else
	{
		bufferofs = word_499C7;
		displayofs = bufferofs + TERMINATORSCREENWIDTH/2;
	}
	

	// Set screen to the display buffer
	VW_SetScreen(displayofs, 0);
	
	// Draw the title screen pic (to the back buffer)
	VW_DrawPic(0,0, TITLEPIC);

	// 
	pageDifference = displayofs - bufferofs;

	asm mov es, screenseg;

	for (i = 0; i < 360; i++)
	{
		int varC, var12;
		int var10 = i - 160;
		if (var10 < 0)
			var10 = 0;

		var12 = i;
		if (var12 >= 200)
			var12 = 199;

		for (varC = var10; varC <= var12; varC++)
		{
			varE = bufferofs + ylookup[varC];
			for (var8 = 0; var8 < 2; var8++)
			{
				int varA = columns1[rows1[varC]];
				if (++rows1[varC] == 320)
					rows1[varC] = 0;

				//loc_14D96
asm				mov     cx, varA;
asm				mov     si, cx;
asm				and     si, 7;
asm				cli;
asm				mov     dx, GC_INDEX;
asm				mov     al, GC_BITMASK;
asm				mov     ah, byte ptr bitmasks[si];
asm				out     dx, ax;
asm				sti;

asm				mov     si, varE;
asm				shr     cx, 1;
asm				shr     cx, 1;
asm				shr     cx, 1;
asm				add     si, cx;
asm				mov     di, si;
asm				add     di, pageDifference;

asm				mov     dx, SC_INDEX;
asm				mov     ax, (1 * 256) + SC_MAPMASK;
asm				out     dx, ax          
asm				mov     dx, GC_INDEX;
asm				mov     ax, (0 * 256) + GC_READMAP;
asm				out     dx, ax        
asm				mov     bl, es:[si];
asm				xchg    bl, es:[di];


asm				mov     dx, SC_INDEX;
asm				mov     ax, (2 * 256) + SC_MAPMASK;
asm				out     dx, ax;
asm				mov     dx, GC_INDEX;
asm				mov     ax, (1 * 256) + GC_READMAP;
asm				out     dx, ax;         
asm				mov     bl, es:[si];
asm				xchg    bl, es:[di];

asm				mov     dx, SC_INDEX;
asm				mov     ax, (4 * 256) + SC_MAPMASK;
asm				out     dx, ax;          
asm				mov     dx, GC_INDEX;
asm				mov     ax, (2 * 256) + GC_READMAP;
asm				out     dx, ax;
asm				mov     bl, es:[si];
asm				xchg    bl, es:[di];

asm				mov     dx, SC_INDEX;
asm				mov     ax, (8 * 256) + SC_MAPMASK;
asm				out     dx, ax; 
asm				mov     dx, GC_INDEX;
asm				mov     ax, (3 * 256) + GC_READMAP; 
asm				out     dx, ax;
asm				mov     bl, es:[si];
asm				xchg    bl, es:[di];


			}

		}

		VW_WaitVBL(1);
		if (IN_IsUserInput())
			if (LastScan == sc_F1)
				LastScan = sc_Space;

		if (LastScan)
		{
			EGABITMASK(0xFF);
			EGAMAPMASK(0x0F);
			return;
		}

	}
	
	EGABITMASK(0xFF);
	EGAMAPMASK(0x0F);
	IN_UserInput(840, false);
}

void CK_DrawTerminator(void)
{
	unsigned i,j;
	unsigned cmdrWidthX100;

	word far *linestart; 
	byte far *vidPtr; // perhaps srcPtr, destPtr, respectively


	word unkptrs[200];


	boolean terminator_complete = false;
	CA_SetAllPurge();

	// Prepare screen with black palette
	colors[0][16] = bordercolor;
	_ES=FP_SEG(&colors[0]);
	_DX=FP_OFF(&colors[0]);
	_AX=0x1002;
	geninterrupt(0x10);

	VW_ClearVideo(0);
	VW_SetLineWidth(TERMINATORSCREENWIDTH);


	// Cache Intro Bitmaps
	CA_CacheGrChunk(TITLEPIC);
	CA_CacheGrChunk(I_COMMANDEREXTERN);
	CA_CacheGrChunk(I_KEENEXTERN);


	ck_introKeen = grsegs[I_KEENEXTERN];
	ck_introCommander = grsegs[I_COMMANDEREXTERN];

	EGAMAPMASK(1);


	// Because "KEEN" needs to end up on the right side of the screen,
	// an extra padding amount is added to the left side of the Keen graphic
	ck_introScreenWidth = ck_introKeen->width + 25 * 8;

	// Set the screen to the right of the "KEEN" graphic
	VW_SetScreen(ck_introScreenWidth/8 + 1, 0);

	// Copy each line of the KEEN graphic into video memory, accounting for the padding amount
	for (i = 0; i < 200; i++)
	{
		linestart = MK_FP(ck_introKeen,  ck_introKeen->linestarts[i]); 
		vidPtr = MK_FP(0xA000 , ylookup[i] + 25);

		TerminatorExpandRLE(linestart, vidPtr);
	}

	// Copy the KEEN graphic to the second page
	VW_ScreenToScreen(0,TERMINATORSCREENWIDTH/2,109,200);

	ck_introCommanderWidth = (ck_introCommander->width+7)>>3;
	ck_introCommanderWidth = (ck_introCommanderWidth+3)&0xFFFE;
	cmdrWidthX100 = ck_introCommanderWidth * 100;


	for (j = 0; j < 8; j++)
		MM_GetPtr(&shiftedCmdrBMPsegs[j],cmdrWidthX100);


	ck_introKeen = grsegs[I_KEENEXTERN];
	ck_introCommander = grsegs[I_COMMANDEREXTERN];


	for (i = 0; i < 100; i++)
	{

		unkptrs[2*i] = unkptrs[2*i+1] = i * ck_introCommanderWidth; 

		linestart = MK_FP(ck_introCommander, ck_introCommander->linestarts[2*i]);
		vidPtr = MK_FP(shiftedCmdrBMPsegs[0],unkptrs[2*i]);

		TerminatorExpandRLE(linestart, vidPtr);

	}


	// loc_150A5
	// Looks like we're making 8 shifts of the "Commander" graphic
	for (j = 1; j < 8; j++)
	{
		memptr var10 = shiftedCmdrBMPsegs[j-1];
		memptr var12 = shiftedCmdrBMPsegs[j];

asm		mov ds, var10;
asm		mov es, var12;
asm		mov cx, cmdrWidthX100;
asm		clc;
asm		xor si, si;
asm		xor di, di;

copyloop:
asm		lodsb;
asm		rcr al, 1;
asm		stosb;
asm		loop copyloop;

asm		mov ax, ss;
asm		mov ds, ax;
	}


	// Set the terminator palette
	terminator_palette2[16] = bordercolor;
	terminator_palette1[16] = bordercolor;

	// and the border.
	_AX = _DS;
	_ES = _AX;
	_DX = (int)&terminator_palette2;
	_AX = 0x1002;
	geninterrupt(0x10);

	for (i = 0; i < 4; i++)
		DoubleSizeTerminatorPics(i);


	// Do the terminator
	ck_currentTermPicStartTime = ck_currentTerminatorCredit = 0;
	ck_termCreditStage = -1;
	AnimateTerminator();


	for (i = 0; i < 4; i++)
		MM_FreePtr(&terminator_pic_memptrs[i]);

	for (j = 0; j < 8; j++)
		MM_FreePtr(&shiftedCmdrBMPsegs[j]);


	// Do the Zooming-out transition to the fizzlefade
	if (!LastScan)
	{
		ZoomOutTerminator();
	}

	// Do the fizzlefade
	if (!LastScan)
	{
		FinishPage();
		terminator_complete = true;
	}

	// Free the COMMANDER and KEEN bitmaps
	MM_SetPurge(&(memptr)grsegs[I_COMMANDEREXTERN], 3);
	MM_SetPurge(&(memptr)grsegs[I_KEENEXTERN], 3);

	// Restore video mode to normal
	VW_ClearVideo(0);
	VW_SetLineWidth(SCREENWIDTH);
	VW_SetDefaultColors();
	RF_FixOfs();
	CA_ClearMarks();

	// Handle any keypresses
	if (!LastScan)
		return;

	// Go to help screen
	if (LastScan == sc_F1)
	{
		HelpScreens();
		return;
	}

	if (!terminator_complete)
	{
		RF_FixOfs();

		// Draw Title Screen to back buffer and set screen there
		CA_CacheGrChunk(TITLEPIC);
		VW_DrawPic(0, 0, TITLEPIC);
		VW_UpdateScreen();
		VW_SetScreen(bufferofs, 0);

		// Wait for input
		IN_Ack();//IN_WaitButton();

		// Stop music if it hasn't already stopped
		StopMusic();

		CA_ClearMarks();

		// TODO: If started with /DEMO PARM
#if 0
		if (DemoSwitch)
		{
			ck_levelState = 5;
			restartgame = D_Normal;
			IN_ClearKeysDown();
			CK_NewGame();
			return;
		}
#endif

	}

	US_ControlPanel();
	if (restartgame)
	{
		ck_levelState = 5;
		return;
	}

	if (loadedgame)
		ck_levelState = 6;
	
}

//===========================================================================
//
// STAR WARS
// 
//===========================================================================

// Dynamically assembles 200 scaler functions, one for each row of the screen, that 
// stretch or compress an input row of pixels by a fixed amount and write these to the screen
//
// The dynamic assembly doesn't seem to be necessary, as the message data doesn't 
// actually affect the assembly process.  

#define	EGAPORTSCREENWIDE	42

// Table that adds a row-displacement when selecting a line of the master text image to draw, 
// given a row on visible screen as the index
// Because the table values are skewed, this causes the text to bunch up at the top of the screen.
word ck_SWScreenRowToMasterRowTrans[200];


// The total height, in rows, of the entire star wars message (before graphic transformations)
unsigned ck_starWarsTotalHeight;

// The table of byte values 
memptr ck_starWarsTable;

// The buffer that holds all of the dynamically assembled instructions
memptr ck_SWDrawFuncSeg;

// The entire, unscaled, starwars text message is drawn to screen and then copied to this seg
memptr ck_starWarsMasterText;



void (* ck_SWCurrentDrawFunc)(void);
void (* ck_SWDrawFuncs[200])(void);


char far ck_starWarsText[] =
"Commander Keen in...\n"
"Source Mod Demo!\n"
"\n"
"Keen finds himself\n"
"in a skeleton mod\n"
"with no one to save\n"
"and little to do.\n"
"\n"
"Will our intrepid hero\n"
"be able to stave off \n"
"boredom, or will a\n"
"capable modder provide\n"
"him with a foe\n"
"to vanquish?\n";

// Generates a table of byte values
// This table somehow translates master text data into maskbytes for drawing pixels on screen
void starwars2(void)
{
	byte far *var6;
	word di, var2, si;


	MM_GetPtr((memptr *)&ck_starWarsTable, 0x4000);
	var6 = MK_FP(ck_starWarsTable, 0);

	for (di = 1; di < 0x100; di <<= 1)
		for (var2 = 1; var2 < 0x100; var2 <<=1)
			for (si = 0; si < 0x100; si++)
				*var6++ = (si & di)? (byte)var2 : 0;

}

void BuildScalers(void)
{
	
	longword varA, var10, swDrawFuncSegSize, var14;
	word var8, si, var2A, var28, var24;
	byte var1D, var1E;

	// The width of the text trapezoid at the bottom of the screen
	// The trapezoid width is decreased by delTrapWidth for each row
	// Units are in 1 / (2 ^ 11) 'ths of a pixel
	long trapezoidWidth;
	int delTrapWidth;

	// This points to the location where machine code is written
	// In the disassembly, this is a register variable stored in ES:DI for
	// the duration of the loop
	word register far *esdi;

	// esdi and drawFuncStart might just be one variable in the original source
	// as drawFuncStart doesn't seem to do anything
	word far *drawFuncStart;

	// the width of the text at this row on screen
	unsigned rowwidth;

	// the screen x coord where the starwars text is drawn
	unsigned rowleftmargin;

	// the offset in the video buffer where this row starts
	unsigned rowofs;

	// the offset, mod 8, of the row left margin
	byte rowmask;

	// loop variable
	int row;





	// Generate the table
	starwars2();

	// Allocate a gigantic buffer for the dynamically assembled instructions
	swDrawFuncSegSize = 190000L;
	MM_GetPtr((memptr *)&ck_SWDrawFuncSeg, swDrawFuncSegSize);

	// Set the pointer that writes the assembled instructions to the 
	// start of the previously allocated buffer
	drawFuncStart = MK_FP(ck_SWDrawFuncSeg, 0);
	esdi = drawFuncStart;

	// Set the dimensions of the text trapezoid
	trapezoidWidth = 0xA0000L; 
	delTrapWidth = 0xB33;

	var10 = 0;
	
	// Generate a dynamically-assembled function for each screen line
	for (row = 199; row >= 0; row--)
	{
		// Draw a strip of grey directly into the screen area
		// (because we've set bufferofs to 0, and HLin draws to bufferofs+ylookup)
		VW_Hlin(0,320,row,1);

		// Save the function address for this line
		drawFuncStart = esdi;
		ck_SWDrawFuncs[row] = (void (*)(void))drawFuncStart;

		// make the row width even
		rowwidth = (trapezoidWidth/2 >> 0xB) << 1;

		// The text scrunching is negatively proportional to the width of the row
		// Smaller row width = more scrunched text
		ck_SWScreenRowToMasterRowTrans[row] = var10 >> 0xB;

		var8 = 0xA8000L/rowwidth;

		var14 = 0;
		var10 += var8;

		// 160 is the center of the screen
		rowleftmargin = 160 - rowwidth/2;
		rowofs = ylookup[row] + rowleftmargin/8;
		rowmask = rowleftmargin & 7;

		// machine code!
		// Remember that the low byte comes first

		*esdi++ = 0xD18C;		// mov cx, ss
		*esdi++ = 0xD08E;		// mov ss, ax
		*esdi++ = 0xC781;		// add di, rowofs
		*esdi++ = rowofs;		// 
		*esdi++ = 0xC030;		// xor al, al

		var2A = 0xFFFF;

		// for each byte that is drawn to the screen
		for (si = 0; si < rowwidth; si ++)
		{

			var24 = var14 >> 0xB;
			var1D = var24 / 8;			// some xbyte
			var1E = var24 & 7;			// and the corresponding mask

			var28 = ((7 - var1E) * 8 + 7 - rowmask) << 8;

			if (var1D != var2A)
			{
				// remember that SS = AX, and [bp+dist] uses SS: as segment
				// When called, [SS:BP] -> MasterTextSeg:masterofs 
				*esdi++ = 0x5E8A;		// mov bl, [bp + var1D]
				*((byte far *)esdi)++ = var1D;  
				var2A = var1D;
			}

				// Assume BH = 0;
			*esdi++ = 0x870A;			// or al, [bx + var28] 
			*esdi++ = var28;
			if (++rowmask == 8)
			{
				rowmask = 0;
				*esdi++ = 0x30AA;			// stosb
				*((byte far *)esdi)++ = 0xC0;		// xor al, al
			}

			var14 += var8;
		}

		if (rowmask)
			*((byte far *)esdi)++ = 0xAA;			// stosb

		*esdi++ = 0xD18E;					// mov ss, cx
		*((byte far *)esdi)++ = 0xCB;				// retf;


		// Normalize esdi
		esdi = MK_FP(FP_SEG(esdi) + (FP_OFF(esdi) >> 4), FP_OFF(esdi) & 0xF);

		// Shrink the scale distance for the next line
		trapezoidWidth -= delTrapWidth;

		// The backdrop picture was drawn to the backbuffer (at 0x8000) in CK_DrawStarWars
		// Copy one row of the backdrop, so the image appears to fill the screen from the bottom
		//  as we prepare the star wars effect
		VW_ScreenToScreen(0x8000 + ylookup[row], ylookup[row], 40, 1); // copy 320 x 1 pixels

		if (LastScan)
			return;
	}
}


// Parses a line of starwars text
// and replaces character codes?
void ParseStarWarsLine(char *line)
{
	char c;

	while (*line)
	{
		c = *line;
		if (c >= 'A' && c <= 'Z')
			c -= 33;
		else if (c >= 'a' && c <= 'z')
			c -= 39;
		else if (c == '.')
			c = 0x54;
		else if (c == ',')
			c = 0x55;
		else if (c == '-')
			c = 0x56;
		else if (c == '"')
			c = 0x57;
		else if (c == ' ')
			c = 0x58;
		else if (c == '!')
			c = 0x59;
		else if (c == '\'')
			c = 0x5A;
		else if (c != '\n')
			c = 0x54;

		*line++ = c;
	}


}

// Draws the Star Wars text
void PrintStarWars(void)
{
	char far *storytext;
	char textbuffer[81];
	char c;

	WindowX = 0;
	WindowW = EGAPORTSCREENWIDE * 8; //336;
	PrintY = 1;		// Save a 1 pixel high blank row at the start of the master text buffer
	bufferofs = 0;
	panadjust = 0;

	storytext = ck_starWarsText;
	ck_starWarsTotalHeight = 0;

	// Draw Each Line using the User manager print routines
	while (*storytext)
	{
		char *si = textbuffer;

		do {
			c = *storytext++;
			*si++ = c;

		} while (c != '\n' && c);


		*si = 0;
		ParseStarWarsLine(textbuffer);

		US_CPrint(textbuffer);

		// Move the back buffer down by one line of text
		bufferofs += ylookup[PrintY];
		ck_starWarsTotalHeight += PrintY;
		PrintY = 0;

	}

	// Get a buffer big enough to hold the graphics data
	// And move into it all the text that has just been drawn
	MM_GetPtr((memptr *)&ck_starWarsMasterText, bufferofs);
	EGAREADMAP(1);
	movedata(screenseg, 0, FP_SEG(ck_starWarsMasterText), 0, bufferofs);

	// reset the back buffer pointer and erase all the text that was drawn to the visible area
	bufferofs = 0;
	VW_Bar(0,0, 320,200, 0);

}

void StarWarsLoop(void)
{
	unsigned scrollDist, newTime, delaytime;
	int masterRowToDraw, row;

	TimeCount = lasttimecount = 0L;
	tics = 0;
	
	EGAWRITEMODE(0);
	EGAMAPMASK(8);

	scrollDist = 0;

	// The scroll will move the entire height of the starwars text + 400 pixels
	while (scrollDist < ck_starWarsTotalHeight + 400)
	{
		// Updating rows from the bottom

		for (row = 199; row >= 0; row--)
		{
			masterRowToDraw = scrollDist - ck_SWScreenRowToMasterRowTrans[row];
			if (masterRowToDraw >= 0 && masterRowToDraw < ck_starWarsTotalHeight)
				// We have more text left to draw
				// Pick the source row from the master text buffer
				masterofs = masterRowToDraw * EGAPORTSCREENWIDE;
			else
				// The text hasn't scrolled this far, or it's scrolled beyond this screen line
				// So draw that one pixel-high blank row that we saved earlier!
				masterofs = 0;

		ck_SWCurrentDrawFunc = ck_SWDrawFuncs[row];

		asm	mov	es, [screenseg];
		asm	mov	di, [terminatorOfs];
		asm	mov	ds, [ck_starWarsTable];
		asm	push	bp;
		asm	mov	bp, ss:masterofs;
		asm	mov	ax, ss:ck_starWarsMasterText;
		asm	xor	bh,bh;

		// For reference, here is the entire function that is called
		// Pseudo-conditional-assembly directives are used

		// mov cx, ss			; save stack
		// mov ss, ax			; SS:[BP] -> start of source row (MasterTextSeg:masterofs)
		// add di, rowofs		; ES:[DI] -> start of drawing area
		// xor al, al


		// FOR (si = 0; si <= rowwidth; si++) ; for each byte drawn to the screen
		// IF (var1D != var2A)
		// mov bl, [bp + var1D]
		// ENDIF

		// or al, [bx + var28] 		; DS:[BX] -> TableSeg:

		// IF (++rowmask == 8)
		// stosb
		// xor al, al
		// ENDIF
		// ENDFOR


		// IF (rowmask)
		// stosb
		// ENDIF


		// mov ss, cx				; restore stack
		// retf;				

		asm	cli;
		asm	call	ss:ck_SWCurrentDrawFunc;
		asm	sti;
		

		asm	pop	bp;
		asm	mov	ax, ss;			// restore stack
		asm	mov	ds, ax;

		}


	// Flip pages
	VW_SetScreen(terminatorOfs, 0);
	terminatorPageOn ^= 1;
	terminatorOfs = terminatorPageOn << 0xF; 	// 2nd page at 0x8000


	tics += TimeCount - lasttimecount;
	lasttimecount = TimeCount;

	if (tics > 20)
		tics = 20;

	// scrollDist += 1;
       	scrollDist += tics/4;

	tics &= 3;

	if (LastScan == sc_F1)
		LastScan = sc_Space;

	if (LastScan)
		return;

	}


}

char ck_starWarsPalette[] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x10, 0x06, 0x07,
	0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x00};

void CK_DrawStarWars(void)
{

	// Set palette to black
	colors[0][16] = bordercolor;
	_ES=FP_SEG(&colors[0]);
	_DX=FP_OFF(&colors[0]);
	_AX=0x1002;
	geninterrupt(0x10);

	// Clear out video memory and set the visible screen offest to 0
	// so that we have a clean area to draw the master text to 
	VW_ClearVideo(0);
	VW_SetLineWidth(EGAPORTSCREENWIDE);
	VW_SetScreen(0,0);
	terminatorOfs = terminatorPageOn = 0;

	CA_SetAllPurge();

	// Draw the entire text message to the master buffer
	CA_CacheGrChunk(STARTFONT+2);
	fontnumber = 2;
	PrintStarWars();
	fontnumber = 0;

	// Cache the bg image and draw it to the back buffer
	CA_CacheGrChunk(STARWARSPIC);
	bufferofs = 0x8000;
	VW_DrawPic(0,0,STARWARSPIC);
	CA_SetAllPurge();

	// Set the starwars palette
	ck_starWarsPalette[16] = bordercolor;
	_ES=FP_SEG(&ck_starWarsPalette);
	_DX=FP_OFF(&ck_starWarsPalette);
	_AX=0x1002;
	geninterrupt(0x10);

	// Generate the compiled scalers
	bufferofs = 0;
	BuildScalers();

	// Do the star wars effect
	if (!LastScan)
	{
		StartMusic(0, true);
		StarWarsLoop();
		StopMusic();
	}

	// Free everything and restore normal video settings
	MM_FreePtr((memptr *)&ck_SWDrawFuncSeg);
	MM_FreePtr((memptr *)&ck_starWarsTable);
	MM_FreePtr((memptr *)&ck_starWarsMasterText);

	VW_ClearVideo(0);

	VW_SetLineWidth(SCREENWIDTH);
	VW_SetDefaultColors();
	RF_FixOfs();
	CA_ClearMarks();

	CK_HandleDemoKeys();

}

//===========================================================================

void CK_ShowTitleScreen()
{
	panadjust = 0;
	CA_CacheGrChunk(TITLEPIC);

	// Draw title pic to back buffer
	VW_DrawPic(0,0,TITLEPIC);
	VW_SetScreen(displayofs, 0);

	// Copy the image from the back buffer to the visible buffer
	VW_ScreenToScreen(bufferofs, displayofs, 42, 224);

	// Wait for a bit
	IN_UserInput(420, false);
	CA_ClearMarks();
	CK_HandleDemoKeys();
}

//===========================================================================
//
// DEMO PLAYBACK
// 
//===========================================================================

void CK_OverlayHighScores(void);

typedef struct {
	unsigned level;
	unsigned size;
	byte data[];
} demotype;


void CK_PlayDemo(int demoNumber)
{
	demotype far *demo;
	int demoChunk;

	CK_NewGame();

	demoChunk = STARTDEMOS + demoNumber;

	CA_CacheGrChunk(demoChunk);
	demo = grsegs[demoChunk];
	ck_gameState.currentLevel = demo->level;
	DemoSize = demo->size; 

	MM_GetPtr((memptr *)&DemoBuffer, DemoSize);
	MM_SetLock(&grsegs[demoChunk], true);
	_fmemcpy(DemoBuffer, &(demo->data), DemoSize);
	MM_FreePtr(&grsegs[demoChunk]);

	IN_StartDemoPlayback(DemoBuffer, DemoSize);

	CK_LoadLevel(true);
	
	if (ck_inHighScores)
		CK_OverlayHighScores();

	CK_PlayLoop();
	IN_StopDemo();

	// We have to get rid of the demo buffer, as if we want to play it
	// again, we need a fresh copy. ID_IN modifies the buffer.
	MM_FreePtr((memptr *)&DemoBuffer);
	VW_FixRefreshBuffer();
	CA_ClearMarks();

	// What should we do after playing the demo?
	CK_HandleDemoKeys();

}

CK_HighScore ck_highScores[8] =
{
	{"Id Software - '91", 10000, 0},
	{"", 10000, 0},
	{"Jason Blochowiak", 10000, 0},
	{"Adrian Carmack", 10000, 0},
	{"John Carmack", 10000, 0},
	{"Tom Hall", 10000, 0},
	{"John Romero", 10000, 0},
	{"", 10000, 0},
};

// Draw the high scores overtop the level
void CK_OverlayHighScores(void)
{
	int entry;
	unsigned oldbuf;

	// Draws the map to the master screen and
	// sets the update matrix for both pages, forcing
	// text to be drawn
	RF_NewPosition (0,0);

	// Set the back buffer to start of the tilemap so that
	// the text drawing routines will draw over the tiles
	oldbuf = bufferofs;
	bufferofs = masterofs;

	fontcolor = 12;

	for (entry = 0; entry < 8; entry++)
	{
		char buf[0x10], *c;
		unsigned w, h;
		// Print the name
		PrintY = entry*16+51;
		PrintX = 40;
		US_Print(ck_highScores[entry].name);

		// Print the score, right aligned in the second
		// column of the table
		// sprintf(buf, "%d", ck_highScores[entry].score);
		ultoa(ck_highScores[entry].score, buf, 10);

		// Convert it to high score numbers?
		for (c = buf; *c; c++)
		{
			*c += 0x51;
		}

		// Right Align it
		VW_MeasurePropString(buf, &w, &h);
		PrintX = (0x118-w);
		US_Print(buf);
	}

	fontcolor = (15);
	bufferofs = oldbuf;
}

// Enter name if a high score has been achieved
extern boolean HighScoresDirty;
void CK_SubmitHighScore(long score, unsigned arg_4)
{

	int entry, entryRank, var4;

	CK_HighScore newHighScore;
	strcpy(newHighScore.name, "");
	newHighScore.score = score;
	newHighScore.arg4 = arg_4;
	


	// Check if this entry made the high scores
	entryRank = -1;
	for (entry = 0; entry < 8; entry++)
	{
		if (ck_highScores[entry].score >= newHighScore.score)
			continue;

		if (newHighScore.score == ck_highScores[entry].score)
		{
			if (ck_highScores[entry].arg4 >= newHighScore.arg4)
				continue;
		}

		// Made it in! 
		// Overwrite the old high score
		memcpy(&ck_highScores[entry], &newHighScore, sizeof(newHighScore));
		entryRank = entry;
		HighScoresDirty = true;

		// Move all of the lower ranks down by one
		for (var4 = 8; --var4 > entry; )
			memcpy(&ck_highScores[var4], &ck_highScores[var4-1], sizeof(newHighScore));

		break;
	}
		


	if (entryRank != -1)
	{
		ck_inHighScores = true;
		ck_gameState.currentLevel = 2;
		CK_LoadLevel(true);
		CK_OverlayHighScores();
		fontcolor = (12);

		// Draw the High scores level to both pages 
		RF_Refresh();
		RF_Refresh();

		PrintY = entry*16 + 51;
		PrintX = 40;
		
		US_LineInput(PrintX, PrintY, ck_highScores[entryRank].name, 0, 1, 0x39, 0x70);

		ck_inHighScores = false;
	}

	fontcolor = (15);
}

// Play the high score level
void CK_DoHighScores(void) 
{
	ck_inHighScores = true;
	IN_ClearKeysDown();
	CK_PlayDemo(1);
	ck_inHighScores = false;
}
