/* Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 * Copyright (C) 1993-2014 Flat Rock Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

// KD_MAIN.C
/*
=============================================================================

							KEEN DREAMS

					An Id Software production

=============================================================================
*/

#include "mem.h"
#include "string.h"

#include "CK_DEF.H"
#pragma hdrstop

/*
=============================================================================

						 LOCAL CONSTANTS

=============================================================================
*/

/*
=============================================================================

						 GLOBAL VARIABLES

=============================================================================
*/

char		str[80],str2[20];
boolean		tedlevel;
unsigned	tedlevelnum;
int		extravbls; // Imported from Catacomb 3-D

/*
=============================================================================

						 LOCAL VARIABLES

=============================================================================
*/

void	DebugMemory (void);
void	TestSprites(void);
int		DebugKeys (void);
void	ShutdownId (void);
void	Quit (char *error);
void	InitGame (void);
void	main (void);

//===========================================================================



//===========================================================================


/*
 * Measure the containing box size of a string that spans multiple lines
 */
void CK_MeasureMultiline(const char *str, unsigned *w, unsigned *h)
{
	char c;
	unsigned x, y;
	char buf[80];
	char *p;

	*h = *w = (unsigned) 0;
	p = buf;	/* must be a local buffer */

	while ( (c = *str++) != 0 )
	{
		*p++ = c;

		if ( c == '\n' || *str == 0 )
		{
			VW_MeasurePropString( buf, &x, &y);

			*h += y;
			if ( *w < x )
				*w = x;

			p = (char *) buf;
			// Shouldn't buf be cleared so that a newline is not read over by
			// VH_MeasurePropString?
		}
	}
}

/*
==========================
=
= ShutdownId
=
= Shuts down all ID_?? managers
=
==========================
*/

void ShutdownId (void)
{
  US_Shutdown ();
  SD_Shutdown ();
  IN_Shutdown ();
  RF_Shutdown ();
  VW_Shutdown ();
  CA_Shutdown ();
  MM_Shutdown ();
}

//===========================================================================

/*
==========================
=
= Quit
=
==========================
*/

void Quit (char *error)
{

	void _seg *textseg;

	if (!error)
	{
		CA_SetAllPurge();
		CA_CacheGrChunk(ANSI_ENDTEXT);
		textseg = grsegs[ANSI_ENDTEXT];
	}

	VW_ClearVideo(0);
	VW_SetLineWidth(40);

	ShutdownId ();
	if (error && *error)
	{
#pragma	warn	-pro
#pragma	warn	-nod
		clrscr();			// we can't include CONIO because of a name conflict
#pragma	warn	+nod
#pragma	warn	+pro
		puts(error);
		puts("\n");
		exit(1);
	}

	/*if (!US_Started)*/
	{
		movedata((int)textseg, 11, 0xb800,0, 0xfa0);
#pragma	warn	-pro
#pragma	warn	-nod
		gotoxy(1, 0x18);	// we can't include CONIO because of a name conflict
#pragma	warn	+nod
#pragma	warn	+pro
		
	}

	exit (0);
}

//===========================================================================

/*
==========================
=
= InitGame
=
= Load a few things right away
=
==========================
*/

#if 0
#include "piracy.h"
#endif

void InitGame (void)
{
	extern void		MML_UseSpace (unsigned segstart, unsigned seglength);
	unsigned	segstart,seglength; // Patch from Catacomb 3-D, see note below
	int i;

	MM_Startup ();


#if 0
	// Handle piracy screen...
	//
	movedata(FP_SEG(PIRACY),(unsigned)PIRACY,0xb800,displayofs,4000);
	while ((bioskey(0)>>8) != sc_Return);
#endif


#if GRMODE == EGAGR
	if (mminfo.mainmem < 335l*1024)
	{
#pragma	warn	-pro
#pragma	warn	-nod
		clrscr();			// we can't include CONIO because of a name conflict
#pragma	warn	+nod
#pragma	warn	+pro
		puts ("There is not enough memory available to play the game reliably.  You can");
		puts ("play anyway, but an out of memory condition will eventually pop up.  The");
		puts ("correct solution is to unload some TSRs or rename your CONFIG.SYS and");
		puts ("AUTOEXEC.BAT to free up more memory.\n");
		puts ("Do you want to (Q)uit, or (C)ontinue?");
		i = bioskey (0);
		if ( (i>>8) != sc_C)
			Quit ("");
	}
#endif

	US_TextScreen();

	VW_Startup ();
	RF_Startup ();
	IN_Startup ();
	SD_Startup ();
	US_Startup ();

	US_UpdateTextScreen();

	CA_Startup ();

//
// Set the cache box callbacks
//
	drawcachebox = CK_BeginCacheBox;
	updatecachebox = CK_UpdateCacheBox; 
	finishcachebox = CK_FinishCacheBox;


	US_Setup ();


//
// load in and lock down some basic chunks
//

	CA_ClearMarks ();

	CA_MarkGrChunk(STARTFONT);
	CA_MarkGrChunk(STARTTILE8);
	CA_MarkGrChunk(STARTTILE8M);
	CA_MarkGrChunk(STATUSLEFTPICM);
	CA_MarkGrChunk(STATUSTOPPICM);

	CA_CacheMarks (NULL);

	MM_SetLock (&grsegs[STARTFONT],true);
	MM_SetLock (&grsegs[STARTTILE8],true);
	MM_SetLock (&grsegs[STARTTILE8M],true);
	MM_SetLock (&grsegs[STATUSLEFTPICM],true);
	MM_SetLock (&grsegs[STATUSTOPPICM],true);


	CA_LoadAllSounds ();

	fontcolor = WHITE;

	US_FinishTextScreen();

#if 0
	VW_SetScreenMode (GRMODE);
	VW_ClearVideo (BLACK);
#endif
	// PATCH: Instead of (just) the above couple of lines, we use this
	// code imported from Catacomb 3-D (along with the ID Engine)
//
// reclaim the memory from the linked in text screen
//
	segstart = FP_SEG(&introscn);
	seglength = 4000/16;
	if (FP_OFF(&introscn))
	{
		segstart++;
		seglength--;
	}

	MML_UseSpace (segstart,seglength);

	VW_SetScreenMode (GRMODE);
	VW_ColorBorder (3);
	VW_ClearVideo (BLACK);

//
// initialize variable(s)
//
	VW_SetLineWidth(SCREENWIDTH);

	US_SetLoadSaveHooks(LoadTheGame,SaveTheGame,CK_ResetGame);
}

/*
=====================
=
= DemoLoop
=
=====================
*/

void
DemoLoop (void)
{
	int demoNumber;

//
// demo loop
//
/*
	 * Commander Keen could be 'launched' from the map editor TED to test a map.
	 * This was implemented by having TED launch keen with the /TEDLEVEL xx
	 * parameter, where xx is the level number.
	 */

	if (tedlevel)
	{
		NoWait = true; // ?

		CK_NewGame();
		CA_LoadAllSounds();
		ck_gameState.currentLevel = tedlevelnum;
		restartgame = D_Normal;

		// TODO: Support selecting difficulty via an extra command line
		// argument ("easy", "normal", "hard")

		CK_GameLoop();
		Quit(0); // run_ted
	}

	/*
	 * Handle "easy" "normal" and "hard" parameters here
	 */

	// Given we're not coming from TED, run through the demos.

	demoNumber = 0;
	ck_levelState = 0;

	while (true)
	{
		switch (demoNumber++)
		{
		case 0:	// Terminator scroller and Title Screen
			// If no pixel panning capability
			// Then the terminator screen isn't shown
#if GRMODE == EGAGR
				CK_DrawTerminator();	//TODO: Move this to an episode struct.
#elif GRMOD == CGAGR
				CK_ShowTitleScreen();
#endif
			break;
		case 1:
			CK_DrawStarWars();
			break;
		case 2:
			CK_PlayDemo(0);
			break;
		case 3:
			// CK_PlayDemo(1);
			break;
		case 4:
			CK_DoHighScores();
			break;
		case 5:
			// CK_PlayDemo(2);
			break;
		case 6:
			// CK_PlayDemo(3);
			CK_HandleDemoKeys();
			demoNumber = 0;
			break;
		}

		// Game Loop
		while (1)
		{
			if (ck_levelState == 5 || ck_levelState == 6)
			{
				CK_GameLoop();
				CK_DoHighScores();
				if (ck_levelState == 5 || ck_levelState == 6)
					continue;

				CK_ShowTitleScreen();

				if (ck_levelState == 5 || ck_levelState == 6)
					continue;
			}
			else
			{
				break;
			}
		}
	}

#pragma warn    -rch
	Quit("Demo loop exited!?");
#pragma warn    +rch
}


//===========================================================================

/*
==========================
=
= main
=
==========================
*/

void main (void)
{
	InitGame();

	DemoLoop();					// DemoLoop calls Quit when everything is done
	Quit("Demo loop exited???");
}

