/* Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 * Copyright (C) 1993-2014 Flat Rock Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

// CK_OBJ.C

#include "CK_DEF.H"
#pragma	hdrstop

// This file contains some object functions (think, etc) which are common to
// several episodes.

/*
=============================================================================

LOCAL CONSTANTS

=============================================================================
*/


/*
=============================================================================

GLOBAL VARIABLES

=============================================================================
*/

CK_action CK_act_item;
CK_action CK_act_item2;

CK_action CK_ACT_VitalinNotify1;
CK_action CK_ACT_VitalinNotify2;
CK_action CK_ACT_VitalinNotify3;
CK_action CK_ACT_StandPlatform;
CK_action CK_act_AxisPlatform;
CK_action CK_ACT_FallPlat0;
CK_action CK_ACT_FallPlat1;
CK_action CK_ACT_FallPlat2;


void CK_Fall(CK_object *obj);
void CK_Fall2(CK_object *obj);
void CK_Glide(CK_object *obj);
void CK_DeadlyCol(CK_object *o1, CK_object *o2);
void CK_DeadlyCol2(CK_object *o1, CK_object *o2);
void CK_BasicDrawFunc1(CK_object *obj);
void CK_BasicDrawFunc2(CK_object *obj);
void CK_BasicDrawFunc4(CK_object *obj);
void CK_StunCreature(CK_object *creature, CK_object *stunner, CK_action *new_creature_act);
void CK_DoorOpen(CK_object *obj);
void CK_SecurityDoorOpen(CK_object *obj);
void CK_SpawnItem(int tileX, int tileY, int itemNumber);
void CK_PointItem(CK_object *obj);
void CK_SpawnStandPlatform(int tileX, int tileY);
void CK_SpawnAxisPlatform(int tileX, int tileY, int direction);
void CK_AxisPlatform(CK_object *obj);
void CK_SpawnFallPlat(int tileX, int tileY);
void CK_FallPlatSit (CK_object *obj);
void CK_FallPlatFall (CK_object *obj);
void CK_FallPlatRise (CK_object *obj);
/*
=============================================================================

LOCAL VARIABLES

=============================================================================
*/


//===========================================================================
//
// GENERIC ACTOR BEHAVIOURS
// 
//===========================================================================

// Think function for adding gravity

void CK_Fall(CK_object *obj)
{
	CK_PhysGravityHigh(obj);
	ck_nextX = obj->velX * tics;
}

// Think function for adding a slightly lower amount of gravity

void CK_Fall2(CK_object *obj)
{
	CK_PhysGravityMid(obj);
	ck_nextX = obj->velX * tics;
}

void CK_Glide(CK_object *obj)
{
	ck_nextX = obj->velX * tics;
	ck_nextY = obj->velY * tics;
}

void CK_ObjBadState (void) {
	Quit("Object with bad state!");
}

void CK_DeadlyCol(CK_object *o1, CK_object *o2)
{
	if (o2->type == CT_Stunner)
	{
		CK_ShotHit(o2);
	}
	else if (o2->type == CT_Player)
	{
		CK_KillKeen();
	}
}

void CK_DeadlyCol2(CK_object *o1, CK_object *o2)
{
	if (o2->type == CT_Player)
	{
		CK_KillKeen();
	}
}


void CK_BasicDrawFunc1(CK_object *obj)
{
	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
}

/*
 * For walking and turning around at edges 
 */
void CK_BasicDrawFunc2(CK_object *obj)
{
	// Hit wall walking right; turn around and go left
	if (obj->xDirection == motion_Right && obj->leftTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = motion_Left;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
		// Hit wall walking left; turn around and go right
	else if (obj->xDirection == motion_Left && obj->rightTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = motion_Right;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
		// Walked off of ledge; turn around
	else if (obj->topTI == 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = -obj->xDirection;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}

	RF_PlaceSprite(&(obj->sde), obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
}

/*
 * Think function for stunned creatures
 */
void CK_BasicDrawFunc4(CK_object *obj)
{
	int starsX, starsY;

	// Handle physics
	if (obj->leftTI || obj->rightTI)
	{
		obj->velX = 0;
	}

	if (obj->bottomTI)
	{
		obj->velY = 0;
	}

	if (obj->topTI)
	{
		obj->velX = obj->velY = 0;
		if (obj->currentAction->next)
		{
			CK_SetAction2(obj, obj->currentAction->next);
		}
	}

	// Draw the primary chunk
	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);


	// Draw the stunner stars, offset based on the initial type of the stunned
	// critter

	starsX = starsY = 0;

	switch (obj->user4)
	{
		// Enemy specific star-offsets
	}

	// Tick the star 3-frame animation forward
	if ((obj->user1 += tics) > 10)
	{
		obj->user1 -= 10;
		if (++obj->user2 >= 3)
			obj->user2 = 0;
	}

	// FIXME: Will cause problems on 64-bit systems
	RF_PlaceSprite((void **)&obj->user3, obj->posX + starsX, obj->posY + starsY, obj->user2 + STUNNEDSTARS1SPR, false, 3);
}

void CK_StunCreature(CK_object *creature, CK_object *stunner, CK_action *new_creature_act)
{
	// Kill the stunner shot
	CK_ShotHit(stunner);

	// Set stunned creature action
	creature->user1 = creature->user2 = creature->user3 = 0;
	creature->user4 = creature->type;
	CK_SetAction2(creature, new_creature_act);
	creature->type = CT_StunnedCreature;

	// Make the creature jump up a bit
	if ((creature->velY -= 0x18) < -0x30)
		creature->velY = -0x30;
}

//===========================================================================
//
// DOORS
// 
//===========================================================================

CK_action CK_ACT_DoorOpen1 = {0, 0, AT_UnscaledOnce, 0, 0, 10, 0, 0, CK_DoorOpen, NULL, NULL, &CK_ACT_DoorOpen2};
CK_action CK_ACT_DoorOpen2 = {0, 0, AT_UnscaledOnce, 0, 0, 10, 0, 0, CK_DoorOpen, NULL, NULL, &CK_ACT_DoorOpen3};
CK_action CK_ACT_DoorOpen3 = {0, 0, AT_UnscaledOnce, 0, 0, 10, 0, 0, CK_DoorOpen, NULL, NULL, NULL};

void CK_DoorOpen(CK_object *obj)
{
	int i;
	unsigned tilesToReplace[0x30];

	if (obj->user1 + 2 > 0x30)
	{
		Quit("Door too tall!");
	}

	for (i = 0; i < obj->user1 + 2; ++i)
	{
		tilesToReplace[i] = CA_TileAtPos(obj->posX, obj->posY+i, 1) + 1;
	}

	RF_MemToMap(tilesToReplace, 1, obj->posX, obj->posY, 1, obj->user1 + 2);
}

void CK_SecurityDoorOpen(CK_object *obj)
{
	int x, y;
	unsigned tilesToReplace[0x30];
	for (y = 0; y < 4; ++y)
	{
		for (x = 0; x < 4; ++x)
		{
			tilesToReplace[y*4+x] = CA_TileAtPos(obj->posX+x, obj->posY+y, 1) - 4;
		}
	}

	RF_MemToMap(tilesToReplace, 1, obj->posX, obj->posY, 4, 4);
	obj->user1++;
	if (obj->user1 == 3)
	{
		obj->currentAction = 0;
	}
}

//===========================================================================
//
// ITEMS 
// 
//===========================================================================

CK_action CK_act_item = {0, 0, AT_UnscaledOnce, 0, 0, 20, 0, 0, CK_PointItem, NULL, CK_BasicDrawFunc1, &CK_act_item2};
CK_action CK_act_item2 = {0, 0, AT_UnscaledOnce, 0, 0, 20, 0, 0, CK_PointItem, NULL, CK_BasicDrawFunc1, &CK_act_item};

CK_action CK_ACT_VitalinNotify1 = {VITALINBURST1SPR, VITALINBURST1SPR, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &CK_ACT_VitalinNotify2};
CK_action CK_ACT_VitalinNotify2 = {VITALINBURST2SPR, VITALINBURST2SPR, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &CK_ACT_VitalinNotify3};
CK_action CK_ACT_VitalinNotify3 = {VITALINBURST3SPR, VITALINBURST3SPR, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, NULL};

//StartSprites + 
unsigned CK_ItemSpriteChunks[] ={
	REDGEM1SPR, YELLOWGEM1SPR, BLUEGEM1SPR, GREENGEM1SPR,
	POINTS100ITEM1SPR, POINTS200ITEM1SPR, POINTS500ITEM1SPR, POINTS1000ITEM1SPR, POINTS2000ITEM1SPR, POINTS5000ITEM1SPR,
	ITEM3UP1SPR, ITEMSTUNNER1SPR, 207
};

void CK_SpawnItem(int tileX, int tileY, int itemNumber)
{

	CK_object *obj = CK_GetNewObj(false);

	obj->clipped = CLIP_not;
	//obj->active = OBJ_ACTIVE;
	obj->zLayer = 2;
	obj->type = CT_Item;	
	obj->posX = tileX << 8;
	obj->posY = tileY << 8;
	obj->yDirection = -1;
	obj->user1 = itemNumber;
	obj->gfxChunk = CK_ItemSpriteChunks[itemNumber];
	obj->user2 = obj->gfxChunk;
	obj->user3 = obj->gfxChunk + 2;
	CK_SetAction(obj, &CK_act_item);

	// TODO: Wrong place to cache?
	/*
	CA_CacheGrChunk(obj->gfxChunk);
	CA_CacheGrChunk(obj->gfxChunk + 1);
	*/
}
void CK_PointItem(CK_object *obj)
{
	//	obj->timeUntillThink = 20;
	obj->visible = true;
	if (++obj->gfxChunk == obj->user3)
		obj->gfxChunk = obj->user2;
}

//===========================================================================
//
// PLATFORMS
// 
//===========================================================================

CK_action CK_ACT_StandPlatform = {PLATFORMSPR, PLATFORMSPR, AT_UnscaledOnce, 0, 0, 32000, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &CK_ACT_StandPlatform};


void CK_SpawnStandPlatform(int tileX, int tileY)
{
	CK_object *obj = CK_GetNewObj(false);

	obj->type = CT_Platform;
	obj->active = OBJ_ACTIVE;
	obj->zLayer = 0;
	obj->posX = tileX << 8;
	obj->posY = obj->user1 = tileY << 8;
	obj->xDirection = 0;
	obj->yDirection = 1;
	obj->clipped = CLIP_not;
	CK_SetAction(obj, &CK_ACT_StandPlatform);
	//obj->gfxChunk = obj->currentAction->chunkLeft;
	CK_ResetClipRects(obj);
}

CK_action CK_act_AxisPlatform = {PLATFORMSPR, PLATFORMSPR, AT_Frame, 0, 0, 0, 0, 0, CK_AxisPlatform, NULL, CK_BasicDrawFunc1, NULL };

void CK_SpawnAxisPlatform(int tileX, int tileY, int direction)
{
	CK_object *obj = CK_GetNewObj(false);

	obj->type = CT_Platform;
	obj->active = OBJ_ALWAYS_ACTIVE;
	obj->zLayer = 0;
	obj->posX = tileX << 8;
	obj->posY = tileY << 8;

	switch (direction)
	{
	case 0:
		obj->xDirection = 0;
		obj->yDirection = -1;
		break;
	case 1:
		obj->xDirection = 1;
		obj->yDirection = 0;
		break;
	case 2:
		obj->xDirection = 0;
		obj->yDirection = 1;
		break;
	case 3:
		obj->xDirection = -1;
		obj->yDirection = 0;
		break;
	}

	CK_SetAction(obj, &CK_act_AxisPlatform);
	//obj->gfxChunk = obj->currentAction->chunkLeft;

	CK_ResetClipRects(obj);
}



void CK_AxisPlatform(CK_object *obj)
{
	unsigned nextPosUnit, nextPosTile;

	if (ck_nextX || ck_nextY)
	{
		return;
	}
	//TODO: Implement properly.
	ck_nextX = obj->xDirection * 12 * tics;
	ck_nextY = obj->yDirection * 12 * tics;

	if (obj->xDirection == 1)
	{
		nextPosUnit = obj->clipRects.unitX2 + ck_nextX;
		nextPosTile = nextPosUnit >> 8;
		if (obj->clipRects.tileX2 != nextPosTile && 
		 *(mapsegs[2] + (mapwidth * obj->clipRects.tileY1 + nextPosTile)) == 0x1F)
		{
			obj->xDirection = -1;
			//TODO: Change DeltaVelocity
			ck_nextX -= (nextPosUnit & 255);
		}
	}
	else if (obj->xDirection == -1)
	{
		nextPosUnit = obj->clipRects.unitX1 + ck_nextX;
		nextPosTile = nextPosUnit >> 8;
		if (obj->clipRects.tileX1 != nextPosTile && 
		 *(mapsegs[2] + (mapwidth * obj->clipRects.tileY1 + nextPosTile)) == 0x1F)
		{
			obj->xDirection = 1;
			//TODO: Change DeltaVelocity
			//CK_PhysUpdateX(obj, 256 - nextPosUnit&255);
			ck_nextX += (256 - nextPosUnit) & 255;
		}
	}
	else if (obj->yDirection == 1)
	{
		nextPosUnit = obj->clipRects.unitY2 + ck_nextY;
		nextPosTile = nextPosUnit >> 8;
		if (obj->clipRects.tileY2 != nextPosTile && 
		 *(mapsegs[2] + (mapwidth * nextPosTile + obj->clipRects.tileX1)) == 0x1F)
		{
			if (CA_TileAtPos(obj->clipRects.tileX1, nextPosTile - 2, 2) == 0x1F)
			{
				//Stop the platform.
				obj->visible = true;
				ck_nextY = 0;
			}
			else
			{
				obj->yDirection = -1;
				//TODO: Change DeltaVelocity
				ck_nextY -= ( nextPosUnit & 255);
			}
		}
	}
	else if (obj->yDirection == -1)
	{
		nextPosUnit = obj->clipRects.unitY1 + ck_nextY;
		nextPosTile = nextPosUnit >> 8;
		if (obj->clipRects.tileY1 != nextPosTile && 
		 *(mapsegs[2] + (mapwidth * nextPosTile + obj->clipRects.tileX1)) == 0x1F)
		{
			if (CA_TileAtPos(obj->clipRects.tileX1, nextPosTile + 2, 2) == 0x1F)
			{
				// Stop the platform.
				obj->visible = true;
				ck_nextY = 0;
			}
			else
			{
				obj->yDirection = 1;
				//TODO: Change DeltaVelocity
				ck_nextY +=  256 - (nextPosUnit & 255);
			}
		}
	}
}
CK_action CK_ACT_FallPlat0 = {PLATFORMSPR, PLATFORMSPR, AT_Frame, 0x0, 0x0, 0x0, 0x0, 0x0, CK_FallPlatSit, NULL, CK_BasicDrawFunc1, NULL};
CK_action CK_ACT_FallPlat1 = {PLATFORMSPR, PLATFORMSPR, AT_Frame, 0x0, 0x0, 0x0, 0x0, 0x0, CK_FallPlatFall, NULL, CK_BasicDrawFunc1, NULL};
CK_action CK_ACT_FallPlat2 = {PLATFORMSPR, PLATFORMSPR, AT_ScaledFrame, 0x0, 0x0, 0x0, 0x0, -32, CK_FallPlatRise, NULL, CK_BasicDrawFunc1, NULL};


void CK_SpawnFallPlat(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Platform;
	new_object->active = OBJ_ALWAYS_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->user1 = new_object->posY = tileY << 8;
	new_object->xDirection = motion_None;
	new_object->yDirection = motion_Down;
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &CK_ACT_FallPlat0);
}

void CK_FallPlatSit (CK_object *obj)
{

	if (obj == ck_keenState.platform)
	{
		ck_nextY = tics * 16;
		obj->velY = 0;
		if ((unsigned)(obj->posY + ck_nextY - obj->user1) >= 0x80)
			obj->currentAction = &CK_ACT_FallPlat1;
	}
}

void CK_FallPlatFall (CK_object *obj)
{
	unsigned newY, newYT;

	CK_PhysGravityHigh(obj);
	newY = obj->clipRects.unitY2 + ck_nextY;
	newYT = newY >> 8;

	// Stop falling if platform hits a block
	if ((obj->clipRects.tileY2 != newYT) && (CA_TileAtPos(obj->clipRects.tileX1, newYT, 2) == 0x1F))
	{
		ck_nextY = 0xFF - (obj->clipRects.unitY2 & 0xFF);
		if (ck_keenState.platform != obj)
			obj->currentAction = &CK_ACT_FallPlat2;
	}
}

void CK_FallPlatRise (CK_object *obj)
{
	if (ck_keenState.platform == obj)
	{
		obj->velY = 0;
		obj->currentAction = &CK_ACT_FallPlat1;
	}
	else if ((unsigned) obj->posY <= (unsigned) obj->user1)
	{
		ck_nextY = obj->user1 - obj->posY;
		obj->currentAction = &CK_ACT_FallPlat0;
	}
}
